"""Utility functions for humanoid reward functions."""

import warnings

import numpy as np

# The value returned by tolerance() at `margin` distance from `bounds` interval.
_DEFAULT_VALUE_AT_MARGIN = 0.1


def _sigmoids(x, value_at_1, sigmoid):
    """Returns 1 when `x` == 0, between 0 and 1 otherwise.

    Args:
      x: A scalar or numpy array.
      value_at_1: A float between 0 and 1 specifying the output when `x` == 1.
      sigmoid: String, choice of sigmoid type.

    Returns:
      A numpy array with values between 0.0 and 1.0.

    Raises:
      ValueError: If not 0 < `value_at_1` < 1, except for `linear`, `cosine` and
        `quadratic` sigmoids which allow `value_at_1` == 0.
      ValueError: If `sigmoid` is of an unknown type.
    """
    if sigmoid in ("cosine", "linear", "quadratic"):
        if not 0 <= value_at_1 < 1:
            raise ValueError(f"`value_at_1` must be nonnegative and smaller than 1, got {value_at_1}.")
    else:
        if not 0 < value_at_1 < 1:
            raise ValueError(f"`value_at_1` must be strictly between 0 and 1, got {value_at_1}.")

    if sigmoid == "gaussian":
        scale = np.sqrt(-2 * np.log(value_at_1))
        return np.exp(-0.5 * (x * scale) ** 2)

    elif sigmoid == "hyperbolic":
        scale = np.arccosh(1 / value_at_1)
        return 1 / np.cosh(x * scale)

    elif sigmoid == "long_tail":
        scale = np.sqrt(1 / value_at_1 - 1)
        return 1 / ((x * scale) ** 2 + 1)

    elif sigmoid == "reciprocal":
        scale = 1 / value_at_1 - 1
        return 1 / (abs(x) * scale + 1)

    elif sigmoid == "cosine":
        scale = np.arccos(2 * value_at_1 - 1) / np.pi
        scaled_x = x * scale
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", message="invalid value encountered in cos")
            cos_pi_scaled_x = np.cos(np.pi * scaled_x)
        return np.where(abs(scaled_x) < 1, (1 + cos_pi_scaled_x) / 2, 0.0)

    elif sigmoid == "linear":
        scale = 1 - value_at_1
        scaled_x = x * scale
        return np.where(abs(scaled_x) < 1, 1 - scaled_x, 0.0)

    elif sigmoid == "quadratic":
        scale = np.sqrt(1 - value_at_1)
        scaled_x = x * scale
        return np.where(abs(scaled_x) < 1, 1 - scaled_x**2, 0.0)

    elif sigmoid == "tanh_squared":
        scale = np.arctanh(np.sqrt(1 - value_at_1))
        return 1 - np.tanh(x * scale) ** 2

    else:
        raise ValueError(f"Unknown sigmoid type {sigmoid!r}.")


def tolerance(x, bounds=(0.0, 0.0), margin=0.0, sigmoid="gaussian", value_at_margin=_DEFAULT_VALUE_AT_MARGIN):
    """Returns 1 when `x` falls inside the bounds, between 0 and 1 otherwise.

    Args:
      x: A scalar or numpy array.
      bounds: A tuple of floats specifying inclusive `(lower, upper)` bounds for
        the target interval. These can be infinite if the interval is unbounded
        at one or both ends, or they can be equal to one another if the target
        value is exact.
      margin: Float. Parameter that controls how steeply the output decreases as
        `x` moves out-of-bounds.
        * If `margin == 0` then the output will be 0 for all values of `x`
          outside of `bounds`.
        * If `margin > 0` then the output will decrease sigmoidally with
          increasing distance from the nearest bound.
      sigmoid: String, choice of sigmoid type. Valid values are: 'gaussian',
         'linear', 'hyperbolic', 'long_tail', 'cosine', 'tanh_squared'.
      value_at_margin: A float between 0 and 1 specifying the output value when
        the distance from `x` to the nearest bound is equal to `margin`. Ignored
        if `margin == 0`.

    Returns:
      A float or numpy array with values between 0.0 and 1.0.

    Raises:
      ValueError: If `bounds[0] > bounds[1]`.
      ValueError: If `margin` is negative.
    """
    lower, upper = bounds
    if lower > upper:
        raise ValueError("Lower bound must be <= upper bound.")
    if margin < 0:
        raise ValueError("`margin` must be non-negative.")

    in_bounds = np.logical_and(lower <= x, x <= upper)
    if margin == 0:
        value = np.where(in_bounds, 1.0, 0.0)
    else:
        d = np.where(x < lower, lower - x, x - upper) / margin
        value = np.where(in_bounds, 1.0, _sigmoids(d, value_at_margin, sigmoid))

    return float(value) if np.isscalar(x) else value



"""Balance task for humanoid robots.

TODO: Not Implemented because of collision detection issues.
"""

from metasim.utils import configclass

from .base_metacfg import HumanoidTaskMetaCfg


@configclass
class BalanceMetaCfg(HumanoidTaskMetaCfg):
    """Balance task for humanoid robots."""

    episode_length = 1000



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class SlideBlockToTargetMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/slide_block_to_target/v2"
    objects = [
        PrimitiveCubeMetaCfg(
            name="block",
            size=[0.05625, 0.05625, 0.05625],
            color=[1.0, 0.0, 0.0],
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="target",
            usd_path="roboverse_data/assets/rlbench/slide_block_to_target/target/usd/target.usd",
            physics=PhysicStateType.GEOM,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class SetTheTableMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 1200
    traj_filepath = "roboverse_data/trajs/rlbench/set_the_table/v2"
    objects = [
        RigidObjMetaCfg(
            name="holder",
            usd_path="roboverse_data/assets/rlbench/set_the_table/holder/usd/holder.usd",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="fork_visual",
            usd_path="roboverse_data/assets/rlbench/set_the_table/fork_visual/usd/fork_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="knife_visual",
            usd_path="roboverse_data/assets/rlbench/set_the_table/knife_visual/usd/knife_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="spoon_visual",
            usd_path="roboverse_data/assets/rlbench/set_the_table/spoon_visual/usd/spoon_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="plate_visual",
            usd_path="roboverse_data/assets/rlbench/set_the_table/plate_visual/usd/plate_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="glass_visual",
            usd_path="roboverse_data/assets/rlbench/set_the_table/glass_visual/usd/glass_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class EmptyDishwasherMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 600
    traj_filepath = "roboverse_data/trajs/rlbench/empty_dishwasher/v2"
    objects = [
        ArticulationObjMetaCfg(
            name="dishwasher",
            usd_path="roboverse_data/assets/rlbench/empty_dishwasher/dishwasher/usd/dishwasher.usd",
        ),
        RigidObjMetaCfg(
            name="dishwasher_plate_visual",
            usd_path="roboverse_data/assets/rlbench/empty_dishwasher/dishwasher_plate_visual/usd/dishwasher_plate_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import ArticulationObjMetaCfg
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_OBJECTS = [
    ArticulationObjMetaCfg(
        name="microwave_frame_resp",
        usd_path="roboverse_data/assets/rlbench/close_microwave/microwave_frame_resp/usd/microwave_frame_resp.usd",
    ),
]


@configclass
class CloseMicrowaveMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/close_microwave/v2"
    objects = _OBJECTS
    # TODO: add checker


@configclass
class OpenMicrowaveMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/open_microwave/v2"
    objects = _OBJECTS
    # TODO: add checker



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class StackWineMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/stack_wine/v2"
    objects = [
        RigidObjMetaCfg(
            name="wine_bottle_visual",
            usd_path="roboverse_data/assets/rlbench/stack_wine/wine_bottle_visual/usd/wine_bottle_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="rack_bottom_visual",
            usd_path="roboverse_data/assets/rlbench/stack_wine/rack_bottom_visual/usd/rack_bottom_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="rack_top_visual",
            usd_path="roboverse_data/assets/rlbench/stack_wine/rack_top_visual/usd/rack_top_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_DIRTS = [
    PrimitiveCubeMetaCfg(
        name=f"dirt{i}",
        size=[0.01, 0.01, 0.01],
        color=[0.46, 0.31, 0.31],
        physics=PhysicStateType.RIGIDBODY,
    )
    for i in range(5)
]


@configclass
class SweepToDustpanMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/sweep_to_dustpan/v2"
    objects = [
        RigidObjMetaCfg(
            name="Dustpan_4",
            usd_path="roboverse_data/assets/rlbench/sweep_to_dustpan/Dustpan_4/usd/Dustpan_4.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="sweep_to_dustpan_broom_visual",
            usd_path="roboverse_data/assets/rlbench/sweep_to_dustpan/sweep_to_dustpan_broom_visual/usd/sweep_to_dustpan_broom_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        PrimitiveCubeMetaCfg(
            name="broom_holder",
            size=[0.1, 0.1, 0.5],
            color=[0.35, 0.48, 0.54],
            physics=PhysicStateType.GEOM,
        ),
    ] + _DIRTS
    # TODO: add checker



"""Humanoid benchmark tasks."""

# ruff: noqa: F401

from .balance_metacfg import BalanceMetaCfg
from .base_metacfg import HumanoidTaskMetaCfg
from .crawl_metacfg import CrawlMetaCfg
from .cube_metacfg import CubeMetaCfg
from .door_metacfg import DoorMetaCfg
from .hurdle_metacfg import HurdleMetaCfg
from .maze_metacfg import MazeMetaCfg
from .package_metacfg import PackageMetaCfg
from .pole_metacfg import PoleMetaCfg
from .powerlift_metacfg import PowerliftMetaCfg
from .push_metacfg import PushMetaCfg
from .run_metacfg import RunMetaCfg
from .sit_metacfg import SitMetaCfg
from .slide_metacfg import SlideMetaCfg
from .spoon_metacfg import SpoonMetaCfg
from .stair_metacfg import StairMetaCfg
from .stand_metacfg import StandMetaCfg
from .walk_metacfg import WalkMetaCfg



from metasim.cfg.objects import PrimitiveSphereMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class ReachTargetMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/reach_target/v2"
    objects = [
        PrimitiveSphereMetaCfg(
            name="target",
            radius=0.025,
            color=[1.0, 0.0, 0.0],
            physics=PhysicStateType.XFORM,
        ),
        PrimitiveSphereMetaCfg(
            name="distractor0",
            radius=0.025,
            color=[1.0, 0.0, 0.5],
            physics=PhysicStateType.XFORM,
        ),
        PrimitiveSphereMetaCfg(
            name="distractor1",
            radius=0.025,
            color=[1.0, 1.0, 0.0],
            physics=PhysicStateType.XFORM,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class StackBlocksMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/stack_blocks/v2"
    objects = [
        RigidObjMetaCfg(
            name="stack_blocks_target_plane",
            usd_path="roboverse_data/assets/rlbench/stack_blocks/stack_blocks_target_plane/usd/stack_blocks_target_plane.usd",
            physics=PhysicStateType.GEOM,
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_target0",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_target1",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_target2",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_target3",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 0.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_distractor0",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 1.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_distractor1",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 1.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_distractor2",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 1.0, 0.0],
        ),
        PrimitiveCubeMetaCfg(
            name="stack_blocks_distractor3",
            physics=PhysicStateType.RIGIDBODY,
            size=[0.05, 0.05, 0.05],
            color=[1.0, 1.0, 0.0],
        ),
    ]
    # TODO: add checker



"""Push cube on table task for humanoid robots."""

from __future__ import annotations

import torch

from metasim.cfg.checkers import PushChecker
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.types import EnvState
from metasim.utils import configclass
from metasim.utils.humanoid_robot_util import left_hand_position

from .base_metacfg import HumanoidTaskMetaCfg


class SuccessReward:
    """Reward function for the push task when the object is pushed to the destination."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the push reward."""
        self._robot_name = robot_name
        if robot_name != "h1" and robot_name != "h1_simple_hand":
            raise ValueError(f"Unknown robot {robot_name} without hand")

    def __call__(self, states: list[EnvState]) -> float:
        """Compute the push reward."""
        state = states[0]
        box_pos = state["metasim_body_object/object"]["pos"]
        dest_pos = state["metasim_site_destination/destination"]["pos"]
        dgoal = torch.norm(box_pos - dest_pos)
        reward_success = float(dgoal < 0.01)
        return reward_success


class GoalDistanceReward:
    """Reward function for the push task when the object is getting close to the destination."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the push reward."""
        self._robot_name = robot_name

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the push reward."""
        state = states[0]
        box_pos = state["metasim_body_object/object"]["pos"]
        dest_pos = state["metasim_site_destination/destination"]["pos"]
        reward_goal = -torch.norm(box_pos - dest_pos)
        return reward_goal


class HandDistanceReward:
    """Reward function for the push task when the hand is getting close to the object."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the push reward."""
        self._robot_name = robot_name
        if robot_name != "h1" and robot_name != "h1_simple_hand":
            raise ValueError(f"Unknown robot {robot_name} without hand")

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the push reward."""
        state = states[0]
        box_pos = state["metasim_body_object/object"]["pos"]
        left_hand_pos = left_hand_position(state, self._robot_name)
        reward_hand = -torch.norm(box_pos - left_hand_pos)
        return reward_hand


@configclass
class PushMetaCfg(HumanoidTaskMetaCfg):
    """Push cube on table task for humanoid robots."""

    episode_length = 500
    objects = [
        RigidObjMetaCfg(
            name="table",
            mjcf_path="roboverse_data/assets/humanoidbench/push/table/mjcf/table.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
        RigidObjMetaCfg(
            name="object",
            mjcf_path="roboverse_data/assets/humanoidbench/push/object/mjcf/object.xml",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="destination",
            mjcf_path="roboverse_data/assets/humanoidbench/push/destination/mjcf/destination.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/humanoidbench/push/v2/initial_state_v2.json"
    checker = PushChecker()
    reward_weights = [1000, 1, 0.1]  # αs, αt, αh
    reward_functions = [SuccessReward, GoalDistanceReward, HandDistanceReward]



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class ScoopWithSpatulaMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/scoop_with_spatula/v2"
    objects = [
        RigidObjMetaCfg(
            name="spatula_visual",
            usd_path="roboverse_data/assets/rlbench/scoop_with_spatula/spatula_visual/usd/spatula_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        PrimitiveCubeMetaCfg(
            name="Cuboid",
            size=[0.02, 0.02, 0.02],
            color=[0.85, 0.85, 1.0],
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from __future__ import annotations

import mujoco_viewer
import numpy as np
import torch
from dm_control import mjcf
from loguru import logger as log

from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.constants import TaskType
from metasim.types import Extra, Obs, Reward, Success, TimeOut

from ..base import BaseSimHandler, EnvWrapper, IdentityEnvWrapper


class MujocoHandler(BaseSimHandler):
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        super().__init__(scenario, num_envs, headless)
        self.headless = headless

        if num_envs > 1:
            raise ValueError("MujocoHandler only supports single envs, please run with --num_envs 1.")

        self._robot = scenario.robot
        self._mujoco_robot_name = None
        self._robot_num_dof = None
        self._robot_path = self._robot.mjcf_path
        self._gravity_compensation = not self._robot.enabled_gravity

        self.viewer = None
        self.cameras = []
        for camera in scenario.cameras:
            self.cameras.append(camera)
        self._episode_length_buf = 0

        self.replay_traj = False
        self.use_task_decimation = False

        # FIXME: hard code decimation for now
        if self.use_task_decimation:
            self.decimation = self.scenario.decimation
        elif self.replay_traj:
            log.warning("Warning: hard coding decimation to 1 for object states")
            self.decimation = 1
        elif self.task is not None and self.task.task_type == TaskType.LOCOMOTION:
            self.decimation = self.scenario.decimation
        else:
            log.warning("Warning: hard coding decimation to 25 for replaying trajectories")
            self.decimation = 25

    def launch(self) -> None:
        super().launch()
        model = self._init_mujoco()
        self.physics = mjcf.Physics.from_mjcf_model(model)
        self.data = self.physics.data

        self.body_names = [self.physics.model.body(i).name for i in range(self.physics.model.nbody)]
        self.robot_body_names = [
            body_name for body_name in self.body_names if body_name.startswith(self._mujoco_robot_name)
        ]

    def _init_mujoco(self) -> mjcf.RootElement:
        object_paths = []
        for obj in self.objects:
            object_paths.append(obj.mjcf_path)

        mjcf_model = mjcf.RootElement()

        ## Optional: Add ground grid
        # mjcf_model.asset.add('texture', name="texplane", type="2d", builtin="checker", width=512, height=512, rgb1=[0.2, 0.3, 0.4], rgb2=[0.1, 0.2, 0.3])
        # mjcf_model.asset.add('material', name="matplane", reflectance="0.", texture="texplane", texrepeat=[1, 1], texuniform=True)

        for camera in self.cameras:
            direction = np.array([
                camera.look_at[0] - camera.pos[0],
                camera.look_at[1] - camera.pos[1],
                camera.look_at[2] - camera.pos[2],
            ])
            direction = direction / np.linalg.norm(direction)
            up = np.array([0, 0, 1])
            right = np.cross(direction, up)
            right = right / np.linalg.norm(right)
            up = np.cross(right, direction)

            camera_params = {
                "pos": f"{camera.pos[0]} {camera.pos[1]} {camera.pos[2]}",
                "mode": "fixed",
                "fovy": camera.vertical_fov,
                "xyaxes": f"{right[0]} {right[1]} {right[2]} {up[0]} {up[1]} {up[2]}",
                "resolution": f"{camera.width} {camera.height}",
            }
            mjcf_model.worldbody.add("camera", name=f"{camera.name}_custom", **camera_params)

        # Add ground grid, light, and skybox
        mjcf_model.asset.add(
            "texture",
            name="texplane",
            type="2d",
            builtin="checker",
            width=512,
            height=512,
            rgb1=[0, 0, 0],
            rgb2=[1.0, 1.0, 1.0],
        )
        mjcf_model.asset.add(
            "material", name="matplane", reflectance="0.2", texture="texplane", texrepeat=[1, 1], texuniform=True
        )
        ground = mjcf_model.worldbody.add(
            "geom",
            type="plane",
            pos="0 0 0",
            size="100 100 0.001",
            quat="1 0 0 0",
            condim="3",
            conaffinity="15",
            material="matplane",
        )
        self.object_body_names = []
        self.mj_objects = {}
        for i, (obj, obj_path) in enumerate(zip(self.objects, object_paths)):
            obj_mjcf = mjcf.from_path(obj_path)
            obj_attached = mjcf_model.attach(obj_mjcf)
            if not obj.fix_base_link:
                obj_attached.add("freejoint")
            self.object_body_names.append(obj_attached.full_identifier)
            self.mj_objects[obj.name] = obj_mjcf

        robot_xml = mjcf.from_path(self._robot_path)
        robot_attached = mjcf_model.attach(robot_xml)
        if self._robot.freejoint:
            robot_attached.add("freejoint")
        self.robot_attached = robot_attached
        self.mj_objects[self._robot.name] = robot_xml
        self._mujoco_robot_name = robot_xml.full_identifier
        return mjcf_model

    def _get_root_state(self, obj, obj_name):
        """Get root position, rotation, velocity for an object or robot."""
        if obj == self._robot:
            if self._robot.freejoint:
                root_joint = self.physics.data.joint(self._mujoco_robot_name)
                return {
                    "pos": torch.tensor(root_joint.qpos[:3], dtype=torch.float32),
                    "rot": torch.tensor(root_joint.qpos[3:7], dtype=torch.float32),
                    "vel": torch.tensor(root_joint.qvel[:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(root_joint.qvel[3:6], dtype=torch.float32),
                }
            else:
                root_body_id = self.physics.model.body(self._mujoco_robot_name).id
                return {
                    "pos": torch.tensor(self.physics.data.xpos[root_body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[root_body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[root_body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[root_body_id][3:6], dtype=torch.float32),
                }
        else:
            model_name = self.mj_objects[obj_name].model + "/"
            try:
                obj_joint = self.physics.data.joint(model_name)
                return {
                    "pos": torch.tensor(obj_joint.qpos[:3], dtype=torch.float32),
                    "rot": torch.tensor(obj_joint.qpos[3:7], dtype=torch.float32),
                    "vel": torch.tensor(obj_joint.qvel[:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(obj_joint.qvel[3:6], dtype=torch.float32),
                }
            except KeyError:
                obj_body_id = self.physics.model.body(model_name).id
                return {
                    "pos": torch.tensor(self.physics.data.xpos[obj_body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[obj_body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[obj_body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[obj_body_id][3:6], dtype=torch.float32),
                }

    def _get_joint_states(self, obj_name):
        """Get joint positions and velocities."""
        joint_states = {"dof_pos": {}, "dof_vel": {}}

        for joint_id in range(self.physics.model.njnt):
            joint_name = self.physics.model.joint(joint_id).name
            if joint_name.startswith(f"{obj_name}/") or (
                obj_name == self._robot.name and joint_name.startswith(self._mujoco_robot_name)
            ):
                clean_joint_name = joint_name.split("/")[-1]
                if clean_joint_name == "":
                    continue
                joint = self.physics.data.joint(joint_name)

                if len(joint.qpos.shape) == 0 or joint.qpos.shape == (1,):
                    joint_states["dof_pos"][clean_joint_name] = joint.qpos.item()
                    joint_states["dof_vel"][clean_joint_name] = joint.qvel.item()
                else:  # free joint
                    joint_states["dof_pos"][clean_joint_name] = torch.tensor(joint.qpos.copy(), dtype=torch.float32)
                    joint_states["dof_vel"][clean_joint_name] = torch.tensor(joint.qvel.copy(), dtype=torch.float32)

        return joint_states

    def _get_actuator_states(self, obj_name):
        """Get actuator states (targets and forces)."""
        actuator_states = {
            "dof_pos_target": {},
            "dof_vel_target": {},
            "dof_torque": {},
        }

        for actuator_id in range(self.physics.model.nu):
            actuator = self.physics.model.actuator(actuator_id)
            if actuator.name.startswith(self._mujoco_robot_name):
                clean_name = actuator.name[len(self._mujoco_robot_name) :]

                actuator_states["dof_pos_target"][clean_name] = float(
                    self.physics.data.ctrl[actuator_id].item()
                )  # Hardcoded to position control
                actuator_states["dof_vel_target"][clean_name] = None
                actuator_states["dof_torque"][clean_name] = float(self.physics.data.actuator_force[actuator_id].item())

        return actuator_states

    def get_states(self, env_ids: list[int] | None = None) -> list[dict]:
        state = {}

        for obj in self.objects + [self._robot]:
            state[obj.name] = self._get_root_state(obj, obj.name)
            state[obj.name].update(self._get_joint_states(obj.name))
            if obj == self._robot:
                state[obj.name].update(self._get_actuator_states(obj.name))

        for body_id in range(self.physics.model.nbody):
            body = self.physics.model.body(body_id)
            if body.name not in [self._mujoco_robot_name] + [obj.name for obj in self.objects]:
                state[f"metasim_body_{body.name}"] = {
                    "pos": torch.tensor(self.physics.data.xpos[body_id], dtype=torch.float32),
                    "rot": torch.tensor(self.physics.data.xquat[body_id], dtype=torch.float32),
                    "vel": torch.tensor(self.physics.data.cvel[body_id][:3], dtype=torch.float32),
                    "ang_vel": torch.tensor(self.physics.data.cvel[body_id][3:6], dtype=torch.float32),
                    "com": torch.tensor(self.physics.data.subtree_com[body_id], dtype=torch.float32),
                    "com_vel": torch.tensor(self.physics.data.subtree_linvel[body_id], dtype=torch.float32),
                }

        for site_id in range(self.physics.model.nsite):
            site = self.physics.model.site(site_id)

            linear_vel = None
            angular_vel = None
            for sensor_id in range(self.physics.model.nsensor):
                sensor = self.physics.model.sensor(sensor_id)
                if self.physics.model.sensor_objid[sensor_id] == site_id:
                    # mjtSensor enum: VELOCIMETER=2, GYRO=3
                    if sensor.type == 2:
                        linear_vel = torch.tensor(
                            self.physics.data.sensordata[sensor_id : sensor_id + 3], dtype=torch.float32
                        )
                    elif sensor.type == 3:
                        angular_vel = torch.tensor(
                            self.physics.data.sensordata[sensor_id : sensor_id + 3], dtype=torch.float32
                        )

            state[f"metasim_site_{site.name}"] = {
                "pos": torch.tensor(self.physics.data.site_xpos[site_id], dtype=torch.float32),
                "rot": torch.tensor(self.physics.data.site_xmat[site_id], dtype=torch.float32),
                "vel": linear_vel if linear_vel is not None else None,
                "ang_vel": angular_vel if angular_vel is not None else None,
            }

        return [state]

    def _set_root_state(self, obj_name, obj_state, zero_vel=False):
        """Set root position and rotation."""
        if "pos" not in obj_state and "rot" not in obj_state:
            return

        if obj_name == self._robot.name:
            if self._robot.freejoint:
                root_joint = self.physics.data.joint(self._mujoco_robot_name)
                root_joint.qpos[:3] = obj_state.get("pos", [0, 0, 0])
                root_joint.qpos[3:7] = obj_state.get("rot", [1, 0, 0, 0])
                if zero_vel:
                    root_joint.qvel[:6] = 0
            else:
                root_body = self.physics.named.model.body_pos[self._mujoco_robot_name]
                root_body_quat = self.physics.named.model.body_quat[self._mujoco_robot_name]
                root_body[:] = obj_state.get("pos", [0, 0, 0])
                root_body_quat[:] = obj_state.get("rot", [1, 0, 0, 0])
        else:
            model_name = self.mj_objects[obj_name].model + "/"
            try:
                obj_joint = self.physics.data.joint(model_name)
                obj_joint.qpos[:3] = obj_state["pos"]
                obj_joint.qpos[3:7] = obj_state["rot"]
                if zero_vel:
                    obj_joint.qvel[:6] = 0
            except KeyError:
                obj_body = self.physics.named.model.body_pos[model_name]
                obj_body_quat = self.physics.named.model.body_quat[model_name]
                obj_body[:] = obj_state["pos"]
                obj_body_quat[:] = obj_state["rot"]

    def _set_joint_state(self, obj_name, obj_state, zero_vel=False):
        """Set joint positions."""
        if "dof_pos" not in obj_state:
            return

        for joint_name, joint_pos in obj_state["dof_pos"].items():
            full_joint_name = (
                f"{self._mujoco_robot_name}{joint_name}" if obj_name == self._robot.name else f"{obj_name}/{joint_name}"
            )
            joint = self.physics.data.joint(full_joint_name)
            joint.qpos = joint_pos
            if zero_vel:
                joint.qvel = 0
            try:
                actuator = self.physics.model.actuator(full_joint_name)
                self.physics.data.ctrl[actuator.id] = joint_pos
            except KeyError:
                pass

    def set_states(self, states, env_ids=None, zero_vel=True):
        super().set_states(states, env_ids)
        if len(states) > 1:
            raise ValueError("MujocoHandler only supports single env state setting")

        for obj_name, obj_state in states[0].items():
            self._set_root_state(obj_name, obj_state, zero_vel)
            self._set_joint_state(obj_name, obj_state, zero_vel)

    def _disable_robotgravity(self):
        gravity_vec = np.array([0.0, 0.0, -9.81])

        self.physics.data.xfrc_applied[:] = 0
        for body_name in self.robot_body_names:
            body_id = self.physics.model.body(body_name).id
            force_vec = -gravity_vec * self.physics.model.body(body_name).mass
            self.physics.data.xfrc_applied[body_id, 0:3] = force_vec
            self.physics.data.xfrc_applied[body_id, 3:6] = 0

    def step(self, action: list[float]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        self._episode_length_buf += 1

        super().step(action)
        joint_targets = action[0]["dof_pos_target"]

        if self.replay_traj:
            for joint_name, target_pos in joint_targets.items():
                joint = self.physics.data.joint(f"{self._mujoco_robot_name}{joint_name}")
                joint.qpos = target_pos
        else:
            for joint_name, target_pos in joint_targets.items():
                actuator = self.physics.data.actuator(f"{self._mujoco_robot_name}{joint_name}")
                actuator.ctrl = target_pos

        if self._gravity_compensation:
            self._disable_robotgravity()
        self.physics.step(self.decimation)

        if self.viewer is not None and self.viewer.is_alive and not self.headless:
            self.viewer.render()

        obs = self.get_observation() if self.headless else None
        success = self.checker.check(self)
        reward = self.get_reward()
        return (
            obs,
            reward,
            success,
            torch.tensor([self._episode_length_buf >= self.scenario.episode_length]),
            {},
        )

    def reset(self, env_ids: list[int] | None = None) -> tuple[Obs, Extra]:
        assert env_ids is None or env_ids == [0], "MujocoHandler only supports single env reset"
        self._episode_length_buf = 0
        self.simulate()
        obs = self.get_observation()
        self.checker.reset(self)
        return obs, {}

    def refresh_render(self) -> None:
        if self.viewer is not None:
            self.physics.forward()
            self.viewer.render()

    def simulate(self):
        if self.viewer is not None:
            self.viewer.close()
        for i in range(self.scenario.task.step_repeat):
            self.physics.forward()
            self.physics.step()
            if not self.headless:
                self.viewer = mujoco_viewer.MujocoViewer(self.physics.model.ptr, self.physics.data.ptr)
                self.viewer.render()

    ############################################################
    ## Utils
    ############################################################
    def render(self):
        if self.headless:
            # Return an image in headless mode
            # TODO: config the camera in both cfg and xml
            return self.physics.render(camera_id=1, width=640, height=480)
        else:
            if self.viewer is not None and self.viewer.is_alive:
                self.viewer.render()
                # Shouldn't return image for non-headless mode, this will block the desktop render loop
                # return self.physics.render(camera_id=1, width=640, height=480)

    def get_observation(self) -> Obs | None:
        obs = {}
        # Shouldn't return image for non-headless mode, this will block the desktop render loop
        if len(self.cameras) == 0 or not self.headless:
            return None
        camera = self.cameras[0]  # Using first camera as specified
        camera_id = "camera_custom"  # XXX: hard code camera id for now

        if "rgb" in camera.data_types:
            rgb_img = self.physics.render(
                width=camera.width,
                height=camera.height,
                camera_id=camera_id,
                depth=False,
            )
            obs["rgb"] = torch.from_numpy(rgb_img.copy()).unsqueeze(0)

        if "depth" in camera.data_types:
            depth_img = self.physics.render(
                width=camera.width,
                height=camera.height,
                camera_id=camera_id,
                depth=True,
            )
            obs["depth"] = torch.from_numpy(depth_img.copy()).unsqueeze(0)

        return obs

    def get_reward(self) -> Reward:
        from metasim.cfg.tasks.base_task_metacfg import BaseRLTaskMetaCfg

        if hasattr(self.task, "reward_fn"):
            return self.task.reward_fn(self.get_states())

        if isinstance(self.task, BaseRLTaskMetaCfg):
            final_reward = torch.zeros(self.num_envs)
            for reward_func, reward_weight in zip(self.task.reward_functions, self.task.reward_weights):
                final_reward += reward_func(self._robot.name)(self.get_states()) * reward_weight
            return final_reward
        else:
            return None

    def get_success(self) -> Success:
        return self.checker.check(self)

    def get_time_out(self) -> TimeOut:
        return torch.tensor([self._episode_length_buf >= self.scenario.episode_length], dtype=torch.bool)

    ############################################################
    ## Misc
    ############################################################
    def get_pos(self, obj_name: str, env_ids: list[int] | None = None) -> torch.FloatTensor:
        return torch.tensor(self.get_states()[0][obj_name]["pos"], dtype=torch.float32).unsqueeze(0)

    @property
    def num_envs(self) -> int:
        return 1

    @property
    def episode_length_buf(self) -> list[int]:
        return [self._episode_length_buf]


MujocoEnv: type[EnvWrapper[MujocoHandler]] = IdentityEnvWrapper(MujocoHandler)



import os

import torch
from loguru import logger as log

from metasim.cfg.lights import BaseLightMetaCfg, CylinderLightMetaCfg, DistantLightMetaCfg
from metasim.cfg.objects import (
    ArticulationObjMetaCfg,
    BaseObjMetaCfg,
    PrimitiveCubeMetaCfg,
    PrimitiveCylinderMetaCfg,
    PrimitiveSphereMetaCfg,
    RigidObjMetaCfg,
)
from metasim.cfg.robots import BaseRobotMetaCfg
from metasim.constants import PhysicStateType

try:
    from .empty_env import EmptyEnv
except:
    pass


def add_object(env: "EmptyEnv", obj: BaseObjMetaCfg) -> None:
    import omni.isaac.lab.sim as sim_utils
    from omni.isaac.lab.assets import (
        Articulation,
        ArticulationCfg,
        RigidObject,
        RigidObjectCfg,
    )

    assert isinstance(obj, BaseObjMetaCfg)
    prim_path = f"/World/envs/env_.*/{obj.name}"
    ## Rigid object
    if isinstance(obj, RigidObjMetaCfg):
        if obj.physics == PhysicStateType.XFORM:
            rigid_props = sim_utils.RigidBodyPropertiesCfg(disable_gravity=True, kinematic_enabled=True)
            collision_props = None
        elif obj.physics == PhysicStateType.GEOM:
            rigid_props = sim_utils.RigidBodyPropertiesCfg(disable_gravity=True, kinematic_enabled=True)
            collision_props = sim_utils.CollisionPropertiesCfg(collision_enabled=True)
        elif obj.physics == PhysicStateType.RIGIDBODY:
            rigid_props = sim_utils.RigidBodyPropertiesCfg()
            collision_props = sim_utils.CollisionPropertiesCfg(collision_enabled=True)
        else:
            raise ValueError(f"Unsupported physics state: {obj.physics}")

        ## Primitive object
        if isinstance(obj, PrimitiveCubeMetaCfg):
            env.scene.rigid_objects[obj.name] = RigidObject(
                RigidObjectCfg(
                    prim_path=prim_path,
                    spawn=sim_utils.MeshCuboidCfg(
                        size=tuple([x * obj.scale for x in obj.size]),
                        mass_props=sim_utils.MassPropertiesCfg(mass=obj.mass),
                        visual_material=sim_utils.PreviewSurfaceCfg(
                            diffuse_color=(obj.color[0], obj.color[1], obj.color[2])
                        ),
                        rigid_props=rigid_props,
                        collision_props=collision_props,
                    ),
                )
            )
            return
        if isinstance(obj, PrimitiveSphereMetaCfg):
            env.scene.rigid_objects[obj.name] = RigidObject(
                RigidObjectCfg(
                    prim_path=prim_path,
                    spawn=sim_utils.MeshSphereCfg(
                        radius=obj.radius,
                        mass_props=sim_utils.MassPropertiesCfg(mass=obj.mass),
                        visual_material=sim_utils.PreviewSurfaceCfg(
                            diffuse_color=(obj.color[0], obj.color[1], obj.color[2])
                        ),
                        rigid_props=rigid_props,
                        collision_props=collision_props,
                    ),
                )
            )
            return
        if isinstance(obj, PrimitiveCylinderMetaCfg):
            env.scene.rigid_objects[obj.name] = RigidObject(
                RigidObjectCfg(
                    prim_path=prim_path,
                    spawn=sim_utils.MeshCylinderCfg(
                        radius=obj.radius,
                        height=obj.height,
                        mass_props=sim_utils.MassPropertiesCfg(mass=obj.mass),
                        visual_material=sim_utils.PreviewSurfaceCfg(
                            diffuse_color=(obj.color[0], obj.color[1], obj.color[2])
                        ),
                        rigid_props=rigid_props,
                        collision_props=collision_props,
                    ),
                )
            )
            return
        ## File-based object
        # scale can be a constant or a tuple, or a list
        if isinstance(obj.scale, (tuple, list)):
            scale = tuple(obj.scale)
        else:
            scale = (obj.scale, obj.scale, obj.scale)

        usd_file_cfg = sim_utils.UsdFileCfg(
            usd_path=obj.usd_path,
            rigid_props=rigid_props,
            collision_props=collision_props,
            scale=scale,
        )
        if isinstance(obj, RigidObjMetaCfg):
            env.scene.rigid_objects[obj.name] = RigidObject(RigidObjectCfg(prim_path=prim_path, spawn=usd_file_cfg))
            return

    ## Articulation object
    if isinstance(obj, ArticulationObjMetaCfg):
        env.scene.articulations[obj.name] = Articulation(
            ArticulationCfg(
                prim_path=prim_path,
                spawn=sim_utils.UsdFileCfg(usd_path=obj.usd_path),
                actuators={},
            )
        )
        return
    raise ValueError(f"Unsupported object type: {type(obj)}")


def add_objects(env: "EmptyEnv", objects: list[BaseObjMetaCfg]) -> None:
    for obj in objects:
        add_object(env, obj)


def add_robot(env: "EmptyEnv", robot: BaseRobotMetaCfg) -> None:
    import omni.isaac.lab.sim as sim_utils
    from omni.isaac.lab.actuators import ImplicitActuatorCfg
    from omni.isaac.lab.assets import Articulation, ArticulationCfg

    cfg = ArticulationCfg(
        spawn=sim_utils.UsdFileCfg(
            usd_path=robot.usd_path,
            rigid_props=sim_utils.RigidBodyPropertiesCfg(),
            articulation_props=sim_utils.ArticulationRootPropertiesCfg(),
        ),
        actuators={
            joint_name: ImplicitActuatorCfg(joint_names_expr=[joint_name], stiffness=None, damping=None)
            for joint_name in robot.actuators
        },
    )
    cfg.prim_path = f"/World/envs/env_.*/{robot.name}"
    cfg.spawn.usd_path = os.path.abspath(robot.usd_path)
    cfg.spawn.rigid_props.disable_gravity = not robot.enabled_gravity
    cfg.spawn.articulation_props.enabled_self_collisions = robot.enabled_self_collisions
    for joint_name, actuator in robot.actuators.items():
        cfg.actuators[joint_name].velocity_limit = actuator.velocity_limit

    robot_inst = Articulation(cfg)
    env.scene.articulations[robot.name] = robot_inst
    env.robot = robot_inst


def _add_light(env: "EmptyEnv", light: BaseLightMetaCfg, prim_path: str) -> None:
    import omni.isaac.lab.sim as sim_utils
    from omni.isaac.lab.sim.spawners import spawn_light

    if isinstance(light, DistantLightMetaCfg):
        spawn_light(
            prim_path,
            sim_utils.DistantLightCfg(
                intensity=light.intensity,
            ),
            orientation=light.quat,
        )
    elif isinstance(light, CylinderLightMetaCfg):
        spawn_light(
            prim_path,
            sim_utils.CylinderLightCfg(
                intensity=light.intensity,
                length=light.length,
                radius=light.radius,
            ),
            translation=light.pos,
            orientation=light.rot,
        )
    else:
        raise ValueError(f"Unsupported light type: {type(light)}")


def add_lights(env: "EmptyEnv", lights: list[BaseLightMetaCfg]) -> None:
    for i, light in enumerate(lights):
        _add_light(env, light, f"/World/envs/env_0/lights/light_{i}")


def get_pose(
    env: "EmptyEnv", obj_name: str, obj_subpath: str | None = None, env_ids: list[int] | None = None
) -> tuple[torch.FloatTensor, torch.FloatTensor]:
    from omni.isaac.core.prims import RigidPrimView

    if env_ids is None:
        env_ids = list(range(env.num_envs))

    if obj_name in env.scene.rigid_objects:
        obj_inst = env.scene.rigid_objects[obj_name]
    elif obj_name in env.scene.articulations:
        obj_inst = env.scene.articulations[obj_name]
    else:
        raise ValueError(f"Object {obj_name} not found")

    if obj_subpath is None:
        pos = obj_inst.data.root_pos_w.cpu()[env_ids] - env.scene.env_origins.cpu()[env_ids]
        rot = obj_inst.data.root_quat_w.cpu()[env_ids]
    else:
        view = RigidPrimView(
            obj_inst.cfg.prim_path + "/" + obj_subpath,
            name=f"{obj_name}_{obj_subpath}_view",
            reset_xform_properties=False,
        )
        pos, rot = view.get_world_poses(indices=env_ids)
        pos = pos - env.scene.env_origins[env_ids]
        pos = pos.cpu()
        rot = rot.cpu()

    assert pos.shape == (len(env_ids), 3)
    assert rot.shape == (len(env_ids), 4)
    return pos, rot


def joint_is_implicit_actuator(joint_name: str, obj_inst) -> bool:
    from omni.isaac.lab.actuators import ImplicitActuatorCfg
    from omni.isaac.lab.assets import Articulation

    assert isinstance(obj_inst, Articulation)
    actuators = [actuator for actuator in obj_inst.actuators.values() if joint_name in actuator.joint_names]
    if len(actuators) == 0:
        log.error(f"Joint {joint_name} could not be found in actuators of {obj_inst.cfg.prim_path}")
        return False
    if len(actuators) > 1:
        log.warning(f"Joint {joint_name} is found in multiple actuators of {obj_inst.cfg.prim_path}")
    actuator = actuators[0]
    return isinstance(actuator, ImplicitActuatorCfg)



from __future__ import annotations

import argparse
import time
from collections import deque

import numpy as np
import rootutils
import wandb
from loguru import logger as log
from rich.logging import RichHandler
from stable_baselines3 import PPO
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.env_util import make_vec_env
from wandb.integration.sb3 import WandbCallback

rootutils.setup_root(__file__, pythonpath=True)
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])

from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.constants import SimType
from metasim.scripts.RL.wrapper_sb3 import MetaSimGymWrapper
from metasim.utils.setup_util import get_robot, get_task


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num_envs", type=int, default=1)
    parser.add_argument("--sim", type=str, default="isaaclab", choices=["isaaclab", "isaacgym", "mujoco"])
    parser.add_argument(
        "--robot",
        type=str,
        default="h1",
        choices=[
            "franka",
            "ur5e_2f85",
            "sawyer",
            "franka_with_gripper_extension",
            "h1_2_without_hand",
            "h1",
            "h1_simple_hand",
        ],
    )
    parser.add_argument("--task", type=str, default="Stand")
    parser.add_argument("--add_table", action="store_true")
    parser.add_argument("--total_timesteps", type=int, default=100000000)
    parser.add_argument("--train_or_eval", type=str, default="train", choices=["train", "eval", "test"])
    parser.add_argument("--model_path", type=str, default=None)
    # PPO specific arguments
    parser.add_argument("--learning_rate", type=float, default=3e-5)
    parser.add_argument("--n_steps", type=int, default=2048)
    parser.add_argument("--batch_size", type=int, default=512)
    parser.add_argument("--n_epochs", type=int, default=10)
    # Wandb arguments
    parser.add_argument("--use_wandb", action="store_true", default=True)
    parser.add_argument("--wandb_project", type=str, default="metasim_rl_training")
    parser.add_argument("--wandb_entity", type=str, default="<your_wandb_entity>")
    args = parser.parse_args()
    return args


class EpisodeLogCallback(BaseCallback):
    """
    Custom callback for plotting additional values in tensorboard.
    """

    def __init__(self, verbose=0):
        super().__init__(verbose)
        self.returns_info = {
            "results/return": [],
            "results/episode_length": [],
            "results/success": [],
            "results/success_subtasks": [],
        }

    def _on_step(self) -> bool:
        infos = self.locals["infos"]
        for idx in range(len(infos)):
            curr_info = infos[idx]
            if "episode" in curr_info:
                self.returns_info["results/return"].append(curr_info["episode"]["r"])
                self.returns_info["results/episode_length"].append(curr_info["episode"]["l"])
                cur_info_success = curr_info.get("success", 0)
                self.returns_info["results/success"].append(cur_info_success)
                cur_info_success_subtasks = curr_info.get("success_subtasks", 0)
                self.returns_info["results/success_subtasks"].append(cur_info_success_subtasks)
        return True

    def _on_rollout_end(self) -> None:
        # Use the model's num_timesteps (cumulative across envs) as the global step.
        global_step = self.model.num_timesteps
        for key in self.returns_info.keys():
            if self.returns_info[key]:
                self.logger.record(key, np.mean(self.returns_info[key]), global_step)
                self.returns_info[key] = []


class EvalCallback(BaseCallback):
    def __init__(self, eval_every: int = 100000, scenario=None, sim_type=None, verbose: int = 0):
        super().__init__(verbose=verbose)
        self.eval_every = eval_every
        self.eval_env = None
        self.scenario = scenario
        self.sim_type = sim_type

    def _on_training_start(self) -> None:
        self.eval_env = MetaSimGymWrapper(scenario=self.scenario, sim_type=self.sim_type, num_envs=1)

    def _on_step(self) -> bool:
        if self.num_timesteps % self.eval_every == 0:
            self.record_video()
        return True

    def record_video(self) -> None:
        print("recording video")
        obs, info = self.eval_env.reset()
        video = [self.eval_env.render().transpose(2, 0, 1)]
        for _ in range(1000):
            action = self.model.predict(obs, deterministic=True)[0]
            obs, _, terminated, truncated, _ = self.eval_env.step(action)
            pixels = self.eval_env.render()
            if pixels is not None:
                video.append(pixels.transpose(2, 0, 1))
            if terminated or truncated:
                break

        if video:
            video = np.stack(video)
            wandb.log({"results/video": wandb.Video(video, fps=10, format="mp4")}, step=self.model.num_timesteps)
        else:
            log.warning("No video recorded")


class LogCallback(BaseCallback):
    """
    Custom callback for plotting additional values in tensorboard.
    """

    def __init__(self, verbose=0, info_keywords=(), args=None):
        super().__init__(verbose)
        self.aux_rewards = {}
        self.aux_returns = {}
        for key in info_keywords:
            self.aux_rewards[key] = np.zeros(args.num_envs)
            self.aux_returns[key] = deque(maxlen=100)

    def _on_step(self) -> bool:
        infos = self.locals["infos"]
        for idx in range(len(infos)):
            for key in self.aux_rewards.keys():
                self.aux_rewards[key][idx] += infos[idx][key]

            if self.locals["dones"][idx]:
                for key in self.aux_rewards.keys():
                    self.aux_returns[key].append(self.aux_rewards[key][idx])
                    self.aux_rewards[key][idx] = 0
        return True

    def _on_rollout_end(self) -> None:
        global_step = self.model.num_timesteps
        for key in self.aux_returns.keys():
            self.logger.record(f"aux_returns_{key}/mean", np.mean(self.aux_returns[key]), global_step)


def main():
    args = parse_args()

    if args.use_wandb:
        run = wandb.init(
            project=args.wandb_project,
            entity=args.wandb_entity,
            config=vars(args),
            sync_tensorboard=True,
            monitor_gym=True,
            save_code=True,
            name=f"SB3-{args.sim}-{args.robot}-{args.task}-{time.strftime('%Y_%m_%d_%H_%M_%S')}",
        )
    else:
        from collections import namedtuple

        Run = namedtuple("Run", ["id"])
        run = Run(id=int(time.time()))

    # Create scenario config
    task = get_task(args.task)
    robot = get_robot(args.robot)
    scenario = ScenarioMetaCfg(task=task, robot=robot, try_add_table=args.add_table, sim=args.sim)

    env = make_vec_env(
        lambda: MetaSimGymWrapper(
            scenario=scenario,
            sim_type=SimType(args.sim),
            num_envs=1,  # Each instance is a single environment
            headless=args.train_or_eval == "train",
        ),
        n_envs=args.num_envs,
    )

    # Initialize PPO algorithm
    model = PPO(
        policy="MlpPolicy",
        env=env,
        learning_rate=args.learning_rate,
        n_steps=args.n_steps,
        batch_size=args.batch_size,
        n_epochs=args.n_epochs,
        verbose=1,
        tensorboard_log=f"./ppo_logs/{run.id}",
        device="cpu",
    )

    # Setup wandb callback with additional monitoring options
    wandb_callback = (
        WandbCallback(
            model_save_freq=100000,
            model_save_path=f"models/{run.id}",
            verbose=2,
            gradient_save_freq=100000,  # Added to track gradients
        )
        if args.use_wandb
        else None
    )

    if args.train_or_eval == "train":
        # Train the agent with additional callbacks
        log.info("Starting training...")
        model.learn(
            total_timesteps=args.total_timesteps,
            log_interval=1,
            callback=[
                EvalCallback(scenario=scenario, sim_type=SimType(args.sim)),
                EpisodeLogCallback(),
                LogCallback(args=args),
            ]
            + ([wandb_callback] if args.use_wandb else []),
            progress_bar=True,
        )

        # Save the trained model
        model_path = f"models/ppo_model_{args.robot}_{args.sim}_{run.id}"
        model.save(model_path)
        log.info(f"Model saved to {model_path}")
    elif args.train_or_eval == "eval":
        # Load the trained model
        model.load(args.model_path)

        # Evaluate the agent
        log.info("Starting evaluation...")
        obs, info = env.reset()
        rewards = []
        for _ in range(1000):
            action, _ = model.predict(obs, deterministic=True)
            obs, reward, terminated, truncated, info = env.step(action)
            rewards.append(reward)
            env.render()
            if terminated or truncated:
                obs, info = env.reset()
                import matplotlib.pyplot as plt

                plt.plot(rewards)
                plt.savefig("rewards.png")
                plt.clf()
                rewards = []

    elif args.train_or_eval == "test":
        # Test the joints
        log.info("Starting evaluation...")
        obs, info = env.reset()
        rewards = []

        step = 0
        target_joint = 0
        while True:
            step += 1
            action = np.array(
                [0] * 19,
                dtype=np.float32,
            )
            action[target_joint] = np.sin(step / 10)
            # action = env.normalize_action(np.array([0,0,-0.4,0.8,-0.4,0,0,-0.4,0.8,-0.4,0,0,0,0,0,0,0,0,0]))
            # action = env.normalize_action(obs[7:26])
            obs, reward, terminated, truncated, info = env.step(action)
            env.render()

    # Close environment and wandb
    env.close()
    wandb.finish()


if __name__ == "__main__":
    main()



"""Sit task for humanoid robots."""

from __future__ import annotations

import torch

from metasim.cfg.checkers import SitChecker
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.types import EnvState
from metasim.utils import configclass, humanoid_reward_util
from metasim.utils.humanoid_robot_util import (
    actuator_forces,
    center_of_mass_position,
    center_of_mass_velocity,
    head_height,
    torso_upright,
)

from .base_metacfg import HumanoidTaskMetaCfg


class SitReward:
    """Reward function for the sit task."""

    def __init__(self, robot_name="h1"):
        """Initialize the sit reward."""
        self.robot_name = robot_name

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the sit reward."""
        results = []
        for state in states:
            # Get robot position (center of mass)
            robot_pos = center_of_mass_position(state, self.robot_name)

            # Get chair position (assuming the chair object is named "sit" based on the config)
            chair_pos = state["metasim_body_chair/chair"]["pos"]

            # Get head height for posture calculation
            head_z = head_height(state, self.robot_name)

            # Get IMU position (assuming it's at the torso)
            imu_z = state[f"metasim_site_{self.robot_name}/imu"]["pos"][2]

            # Get velocity for stillness calculation
            velocity = center_of_mass_velocity(state, self.robot_name)
            vx, vy = velocity[0], velocity[1]

            # Calculate sitting components using tolerance function
            # sittingx = tol(xrobot−xchair, (−0.19, 0.19), 0.2)
            sittingx = humanoid_reward_util.tolerance(robot_pos[0] - chair_pos[0], bounds=(-0.19, 0.19), margin=0.2)

            # sittingy = tol(yrobot−ychair, (0, 0), 0.1)
            sittingy = humanoid_reward_util.tolerance(robot_pos[1] - chair_pos[1], bounds=(0, 0), margin=0.1)

            # sittingz = tol(zrobot, (0.68, 0.72), 0.2)
            sittingz = humanoid_reward_util.tolerance(robot_pos[2], bounds=(0.68, 0.72), margin=0.2)

            # posture = tol(zhead−zIMU, (0.35, 0.45), 0.3)
            posture = humanoid_reward_util.tolerance(head_z - imu_z, bounds=(0.35, 0.45), margin=0.3)

            # stillx = tol(vx, (0, 0), 2)
            stillx = humanoid_reward_util.tolerance(vx, bounds=(0, 0), margin=2)

            # stilly = tol(vy, (0, 0), 2)
            stilly = humanoid_reward_util.tolerance(vy, bounds=(0, 0), margin=2)

            # Get upright component
            upright = humanoid_reward_util.tolerance(
                torso_upright(state, self.robot_name),
                bounds=(0.9, 1.0),
                sigmoid="linear",
                margin=0.2,
                value_at_margin=0,
            )

            # Small control
            small_control = humanoid_reward_util.tolerance(
                actuator_forces(state, self.robot_name),
                margin=10,
                value_at_margin=0,
                sigmoid="quadratic",
            ).mean()
            small_control = (4 + small_control) / 5

            # Calculate combined reward
            # R(s,a) = ((0.5·sittingz + 0.5·sittingx×sittingy)×upright×posture)×e×mean(stillx,stilly)
            sitting_position = 0.5 * sittingz + 0.5 * sittingx * sittingy

            # Fixed: Use a simple average instead of torch.mean() on scalar values
            stillness = (stillx + stilly) / 2.0

            reward = sitting_position * upright * posture * small_control * stillness

            results.append(reward)

        return torch.tensor(results)


@configclass
class SitMetaCfg(HumanoidTaskMetaCfg):
    """Sit task for humanoid robots."""

    episode_length = 1000
    objects = [
        RigidObjMetaCfg(
            name="sit",
            mjcf_path="roboverse_data/assets/humanoidbench/sit/chair/mjcf/chair.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/humanoidbench/sit/v2/initial_state_v2.json"
    checker = SitChecker()
    reward_weights = [1.0]
    reward_functions = [SitReward]



from __future__ import annotations

import numpy as np
import torch
from loguru import logger as log
from pyrep import PyRep
from pyrep.objects.object import Object
from pyrep.robots.arms.panda import Panda
from pyrep.robots.end_effectors.panda_gripper import PandaGripper
from rlbench.backend.robot import Robot

from metasim.types import Action, EnvState, Extra, Obs, Reward, Success, TimeOut
from metasim.utils import to_snake_case

from ..base import BaseSimHandler

# TODO: try best to be independent from RLBench

## Joint names correspondence dict
jd = {
    "franka": {
        "Panda_joint1": "panda_joint1",
        "Panda_joint2": "panda_joint2",
        "Panda_joint3": "panda_joint3",
        "Panda_joint4": "panda_joint4",
        "Panda_joint5": "panda_joint5",
        "Panda_joint6": "panda_joint6",
        "Panda_joint7": "panda_joint7",
        "Panda_gripper_joint1": "panda_finger_joint1",
        "Panda_gripper_joint2": "panda_finger_joint2",
    }
}


class PyrepHandler(BaseSimHandler):
    def launch(self) -> None:
        self.sim = PyRep()
        self.sim.launch("third_party/rlbench/rlbench/task_design.ttt", headless=False)  # TODO: set headless from cfg
        # self.sim.set_simulation_timestep(1 / 60)  # Control frequency 60Hz, TODO: set control frequency from cfg
        task_name = to_snake_case(self.task.__class__.__name__.replace("MetaCfg", ""))
        base_object = self.sim.import_model(f"third_party/rlbench/rlbench/task_ttms/{task_name}.ttm")

        self.arm, self.gripper = Panda(), PandaGripper()
        self.robot_inst = Robot(self.arm, self.gripper)
        self.sim.start()
        self.sim.step()
        # TODO: initialize rlbench scene and task

    ############################################################
    ## Gymnasium main methods
    ############################################################
    def step(self, action: list[Action]) -> tuple[Obs, Reward, Success, TimeOut, Extra]:
        action = action[0]["dof_pos_target"]
        arm_action = [action[jd[self.robot.name][joint.get_name()]] for joint in self.arm.joints]
        gripper_action = [action[jd[self.robot.name][joint.get_name()]] for joint in self.gripper.joints]
        self.arm.set_joint_target_positions(arm_action)
        self.gripper.set_joint_target_positions(gripper_action)
        self.sim.step()
        self.arm.set_joint_target_positions(self.arm.get_joint_positions())
        self.gripper.set_joint_target_positions(self.gripper.get_joint_positions())
        return None, None, torch.tensor([False]), torch.tensor([False]), None

    def reset(self) -> tuple[Obs, Extra]:
        return None, None

    def render(self) -> None:
        pass

    def close(self) -> None:
        self.sim.stop()
        self.sim.shutdown()

    ############################################################
    ## Set states
    ############################################################
    def set_states(self, states: list[EnvState]) -> None:
        assert len(states) == 1  # PyRep only supports one env
        state = states[0]

        ## Set robot
        joint_pos = np.zeros(len(self.arm.joints))
        for i, joint_inst in enumerate(self.arm.joints):
            joint_name = joint_inst.get_name()
            joint_pos[i] = state[self.robot.name]["dof_pos"][joint_name]
        self.arm.set_joint_positions(joint_pos[:7])
        self.gripper.set_joint_positions(joint_pos[7:])
        # TODO: set robot pose #!

        ## Set objects
        for obj in self.objects:
            obj_inst = Object.get_object(obj.name)
            obj_inst.set_position(state[obj.name]["pos"].numpy())
            obj_inst.set_orientation(state[obj.name]["rot"].numpy())

            log.debug("Set object", obj.name, "to", state[obj.name]["pos"], state[obj.name]["rot"])

    ############################################################
    ## Get states
    ############################################################
    def get_states(self) -> list[EnvState]:
        pass



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_CUP_HOLDER = RigidObjMetaCfg(
    name="place_cups_holder_base",
    usd_path="roboverse_data/assets/rlbench/place_cups/place_cups_holder_base/usd/place_cups_holder_base.usd",
    physics=PhysicStateType.XFORM,
)

_CUPS = [
    RigidObjMetaCfg(
        name=f"mug_visual{i}",
        usd_path="roboverse_data/assets/rlbench/place_cups/mug_visual1/usd/mug_visual1.usd",  # reuse same asset
        physics=PhysicStateType.RIGIDBODY,
    )
    for i in range(4)
]


@configclass
class PlaceCupsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/place_cups/v2"
    objects = [_CUP_HOLDER] + _CUPS
    # TODO: add checker


@configclass
class RemoveCupsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/remove_cups/v2"
    objects = [_CUP_HOLDER] + _CUPS
    # TODO: add checker



from __future__ import annotations

from typing import Literal

try:
    import isaacgym  # noqa: F401
except ImportError:
    pass

import imageio
import rootutils
import torch
import tyro
from loguru import logger as log
from rich.logging import RichHandler

rootutils.setup_root(__file__, pythonpath=True)
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])


from metasim.cfg.cameras import PinholeCameraMetaCfg
from metasim.cfg.objects import PrimitiveCubeMetaCfg
from metasim.cfg.randomization import RandomizationMetaCfg
from metasim.cfg.render import RenderMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.constants import PhysicStateType, SimType
from metasim.utils import configclass
from metasim.utils.setup_util import get_sim_handler_class


@configclass
class Args:
    robot: str = "franka"
    scene: str | None = None
    render: RenderMetaCfg = RenderMetaCfg()
    random: RandomizationMetaCfg = RandomizationMetaCfg()

    ## Handlers
    sim: Literal["isaaclab", "isaacgym", "genesis", "pyrep", "pybullet", "sapien", "mujoco", "blender"] = "isaaclab"
    renderer: Literal["isaaclab", "isaacgym", "genesis", "pyrep", "pybullet", "sapien", "mujoco", "blender"] | None = (
        None
    )

    ## Others
    num_envs: int = 1
    try_add_table: bool = True
    object_states: bool = False
    split: Literal["train", "val", "test", "all"] = "all"
    headless: bool = False

    ## Only in args
    save_image_dir: str | None = "tmp"
    save_video_path: str | None = None
    stop_on_runout: bool = False

    def __post_init__(self):
        log.info(f"Args: {self}")


args = tyro.cli(Args)

camera = PinholeCameraMetaCfg(
    name="camera",
    data_types=["rgb", "depth"],
    width=256,
    height=256,
    pos=(1.5, -1.5, 1.5),  # near
    # pos=(12.0, -12.0, 20.0),  # far
    look_at=(0.0, 0.0, 0.0),
)
scenario = ScenarioMetaCfg(
    robot=args.robot,
    try_add_table=False,
    sim=args.sim,
    cameras=[camera],
    headless=args.headless,
)
scenario.objects = [
    PrimitiveCubeMetaCfg(
        name="cube",
        size=(0.1, 0.1, 0.1),
        color=(1.0, 0.0, 0.0),
        physics=PhysicStateType.RIGIDBODY,
        mjcf_path="roboverse_data/assets/maniskill/cube/cube.mjcf",
    )
]

handler_class = get_sim_handler_class(SimType(args.sim))
handler = handler_class(scenario, 1, headless=args.headless)
handler.launch()

handler.set_states([
    {
        "cube": {
            "pos": torch.tensor([0.3, 0.3, 0.05]),
            "rot": torch.tensor([1.0, 0.0, 0.0, 0.0]),
        },
        "franka": {
            "pos": torch.tensor([0.0, 0.0, 0.0]),
            "rot": torch.tensor([1.0, 0.0, 0.0, 0.0]),
            "dof_pos": {
                "panda_joint1": 0.0,
                "panda_joint2": 0.0,
                "panda_joint3": 0.0,
                "panda_joint4": 0.0,
                "panda_joint5": 0.0,
                "panda_joint6": 0.0,
                "panda_joint7": 0.0,
                "panda_finger_joint1": 0.0,
                "panda_finger_joint2": 0.0,
            },
        },
    }
])
if args.sim == "isaaclab":
    obs, _ = handler.reset()
else:
    handler.simulate()
    obs = handler.get_observation()

imageio.imwrite("test.png", obs["rgb"][0])



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class ScrewNailMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/screw_nail/v2"
    objects = [
        ArticulationObjMetaCfg(
            name="block",
            usd_path="roboverse_data/assets/rlbench/screw_nail/block/usd/block.usd",
        ),
        RigidObjMetaCfg(
            name="screw_driver",
            usd_path="roboverse_data/assets/rlbench/screw_nail/screw_driver/usd/screw_driver.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from __future__ import annotations

from metasim.cfg.robots.base_robot_metacfg import BaseRobotMetaCfg
from metasim.cfg.scenario import ScenarioMetaCfg
from metasim.cfg.tasks.base_task_metacfg import BaseTaskMetaCfg

from ..base import BaseSimHandler, EnvWrapper, IdentityEnvWrapper

pass
################################################################################
## Common imports ends here, below are simulator specific imports
################################################################################


import genesis as gs
import numpy as np


class GenesisHandler(BaseSimHandler):
    def __init__(self, scenario: ScenarioMetaCfg, num_envs: int = 1, headless: bool = False):
        super().__init__(scenario, num_envs, headless)
        self._task: BaseTaskMetaCfg = scenario.task
        self._robot: BaseRobotMetaCfg = scenario.robot
        self._robot_init_pose = (0, 0, 0) if not self._robot.default_position else self._robot.default_position

        self._cameras = scenario.cameras
        self._env_ids = np.arange(num_envs)

    def launch(self) -> None:
        super().launch()

        gs.init(backend=gs.cpu)

        scene = gs.Scene(show_viewer=True)
        plane = scene.add_entity(gs.morphs.Plane())
        franka = scene.add_entity(
            gs.morphs.MJCF(file="xml/franka_emika_panda/panda.xml"),
        )

        scene.build()

        for i in range(1000):
            scene.step()


GenesisEnv: type[EnvWrapper[GenesisHandler]] = IdentityEnvWrapper(GenesisHandler)



# ruff: noqa: F401

from .arnold_pickup_object import (
    ArnoldPickupObject3380010MetaCfg,
    ArnoldPickupObject3380020MetaCfg,
    ArnoldPickupObject3380030MetaCfg,
    ArnoldPickupObject3380040MetaCfg,
    ArnoldPickupObject3398110MetaCfg,
    ArnoldPickupObject3398120MetaCfg,
    ArnoldPickupObject3398130MetaCfg,
    ArnoldPickupObject3398140MetaCfg,
    ArnoldPickupObject3398210MetaCfg,
    ArnoldPickupObject3398220MetaCfg,
    ArnoldPickupObject3398230MetaCfg,
    ArnoldPickupObject3398240MetaCfg,
    ArnoldPickupObject3398310MetaCfg,
    ArnoldPickupObject3398320MetaCfg,
    ArnoldPickupObject3398330MetaCfg,
    ArnoldPickupObject3398340MetaCfg,
    ArnoldPickupObject3398410MetaCfg,
    ArnoldPickupObject3398420MetaCfg,
    ArnoldPickupObject3398430MetaCfg,
    ArnoldPickupObject3398510MetaCfg,
    ArnoldPickupObject3398520MetaCfg,
    ArnoldPickupObject3398530MetaCfg,
    ArnoldPickupObject3398540MetaCfg,
    ArnoldPickupObject3398610MetaCfg,
    ArnoldPickupObject3398620MetaCfg,
    ArnoldPickupObject3398630MetaCfg,
    ArnoldPickupObject3398640MetaCfg,
    ArnoldPickupObject3517710MetaCfg,
    ArnoldPickupObject3517720MetaCfg,
    ArnoldPickupObject3517730MetaCfg,
    ArnoldPickupObject3517740MetaCfg,
    ArnoldPickupObject3517810MetaCfg,
    ArnoldPickupObject3517820MetaCfg,
    ArnoldPickupObject3517830MetaCfg,
    ArnoldPickupObject3517840MetaCfg,
    ArnoldPickupObject3517910MetaCfg,
    ArnoldPickupObject3517920MetaCfg,
    ArnoldPickupObject3517930MetaCfg,
    ArnoldPickupObject3517940MetaCfg,
    ArnoldPickupObject3558010MetaCfg,
    ArnoldPickupObject3558020MetaCfg,
    ArnoldPickupObject3558030MetaCfg,
    ArnoldPickupObject3558040MetaCfg,
    ArnoldPickupObject3571010MetaCfg,
    ArnoldPickupObject3571020MetaCfg,
    ArnoldPickupObject3571030MetaCfg,
    ArnoldPickupObject3571040MetaCfg,
    ArnoldPickupObject3574010MetaCfg,
    ArnoldPickupObject3574020MetaCfg,
    ArnoldPickupObject3574030MetaCfg,
    ArnoldPickupObject3593010MetaCfg,
    ArnoldPickupObject3593020MetaCfg,
    ArnoldPickupObject3593030MetaCfg,
    ArnoldPickupObject35171010MetaCfg,
    ArnoldPickupObject35171020MetaCfg,
    ArnoldPickupObject35171030MetaCfg,
    ArnoldPickupObject35171040MetaCfg,
    ArnoldPickupObject35171110MetaCfg,
    ArnoldPickupObject35171120MetaCfg,
    ArnoldPickupObject35171130MetaCfg,
    ArnoldPickupObject35171140MetaCfg,
    ArnoldPickupObject35171210MetaCfg,
    ArnoldPickupObject35171220MetaCfg,
    ArnoldPickupObject35171230MetaCfg,
    ArnoldPickupObject35171240MetaCfg,
    ArnoldPickupObject35171310MetaCfg,
    ArnoldPickupObject35171320MetaCfg,
    ArnoldPickupObject35171330MetaCfg,
    ArnoldPickupObject35171410MetaCfg,
    ArnoldPickupObject35171420MetaCfg,
    ArnoldPickupObject35171430MetaCfg,
    ArnoldPickupObject35171440MetaCfg,
    ArnoldPickupObject35171510MetaCfg,
    ArnoldPickupObject35171520MetaCfg,
    ArnoldPickupObject35171530MetaCfg,
    ArnoldPickupObject35171540MetaCfg,
    ArnoldPickupObject35171610MetaCfg,
    ArnoldPickupObject35171620MetaCfg,
    ArnoldPickupObject35171630MetaCfg,
    ArnoldPickupObject35171640MetaCfg,
    ArnoldPickupObject35171710MetaCfg,
    ArnoldPickupObject35171720MetaCfg,
    ArnoldPickupObject35171730MetaCfg,
    ArnoldPickupObject35171810MetaCfg,
    ArnoldPickupObject35171820MetaCfg,
    ArnoldPickupObject35171830MetaCfg,
    ArnoldPickupObject35191910MetaCfg,
    ArnoldPickupObject35191920MetaCfg,
    ArnoldPickupObject35191930MetaCfg,
    ArnoldPickupObject35191940MetaCfg,
    ArnoldPickupObject35192010MetaCfg,
    ArnoldPickupObject35192020MetaCfg,
    ArnoldPickupObject35192030MetaCfg,
    ArnoldPickupObject35192110MetaCfg,
    ArnoldPickupObject35192120MetaCfg,
    ArnoldPickupObject35192130MetaCfg,
    ArnoldPickupObject35192140MetaCfg,
    ArnoldPickupObject35192210MetaCfg,
    ArnoldPickupObject35192220MetaCfg,
    ArnoldPickupObject35192230MetaCfg,
    ArnoldPickupObject35192310MetaCfg,
    ArnoldPickupObject35192320MetaCfg,
    ArnoldPickupObject35192330MetaCfg,
    ArnoldPickupObject35192340MetaCfg,
    ArnoldPickupObject35192410MetaCfg,
    ArnoldPickupObject35192420MetaCfg,
    ArnoldPickupObject35192430MetaCfg,
    ArnoldPickupObject35192440MetaCfg,
    ArnoldPickupObject35192510MetaCfg,
    ArnoldPickupObject35192520MetaCfg,
    ArnoldPickupObject35192530MetaCfg,
    ArnoldPickupObject35192540MetaCfg,
    ArnoldPickupObject35192610MetaCfg,
    ArnoldPickupObject35192620MetaCfg,
    ArnoldPickupObject35192630MetaCfg,
    ArnoldPickupObject35192640MetaCfg,
    ArnoldPickupObject35202710MetaCfg,
    ArnoldPickupObject35202720MetaCfg,
    ArnoldPickupObject35202730MetaCfg,
    ArnoldPickupObject35202810MetaCfg,
    ArnoldPickupObject35202820MetaCfg,
    ArnoldPickupObject35202830MetaCfg,
    ArnoldPickupObject35202840MetaCfg,
    ArnoldPickupObject35202910MetaCfg,
    ArnoldPickupObject35202920MetaCfg,
    ArnoldPickupObject35202930MetaCfg,
    ArnoldPickupObject35202940MetaCfg,
    ArnoldPickupObject35203010MetaCfg,
    ArnoldPickupObject35203020MetaCfg,
    ArnoldPickupObject35203030MetaCfg,
    ArnoldPickupObject35203040MetaCfg,
    ArnoldPickupObject35203110MetaCfg,
    ArnoldPickupObject35203120MetaCfg,
    ArnoldPickupObject35203130MetaCfg,
    ArnoldPickupObject35203140MetaCfg,
    ArnoldPickupObject35203210MetaCfg,
    ArnoldPickupObject35203220MetaCfg,
    ArnoldPickupObject35203230MetaCfg,
    ArnoldPickupObject35203240MetaCfg,
    ArnoldPickupObject35203310MetaCfg,
    ArnoldPickupObject35203320MetaCfg,
    ArnoldPickupObject35203330MetaCfg,
    ArnoldPickupObject35203340MetaCfg,
    ArnoldPickupObject35203410MetaCfg,
    ArnoldPickupObject35203420MetaCfg,
    ArnoldPickupObject35203430MetaCfg,
    ArnoldPickupObject35203440MetaCfg,
    ArnoldPickupObject35203510MetaCfg,
    ArnoldPickupObject35203520MetaCfg,
    ArnoldPickupObject35203530MetaCfg,
    ArnoldPickupObject35203540MetaCfg,
    ArnoldPickupObject35203610MetaCfg,
    ArnoldPickupObject35203620MetaCfg,
    ArnoldPickupObject35203630MetaCfg,
    ArnoldPickupObject35203640MetaCfg,
    ArnoldPickupObject35743710MetaCfg,
    ArnoldPickupObject35743720MetaCfg,
    ArnoldPickupObject35743730MetaCfg,
    ArnoldPickupObject35743740MetaCfg,
    ArnoldPickupObject35743810MetaCfg,
    ArnoldPickupObject35743820MetaCfg,
    ArnoldPickupObject35743830MetaCfg,
    ArnoldPickupObject35743840MetaCfg,
    ArnoldPickupObject35743910MetaCfg,
    ArnoldPickupObject35743920MetaCfg,
    ArnoldPickupObject35743930MetaCfg,
    ArnoldPickupObject35743940MetaCfg,
    ArnoldPickupObject35744010MetaCfg,
    ArnoldPickupObject35744020MetaCfg,
    ArnoldPickupObject35744030MetaCfg,
    ArnoldPickupObject35744040MetaCfg,
    ArnoldPickupObject35744110MetaCfg,
    ArnoldPickupObject35744120MetaCfg,
    ArnoldPickupObject35744130MetaCfg,
    ArnoldPickupObject35744140MetaCfg,
    ArnoldPickupObject35744210MetaCfg,
    ArnoldPickupObject35744220MetaCfg,
    ArnoldPickupObject35744230MetaCfg,
    ArnoldPickupObject35744240MetaCfg,
    ArnoldPickupObject35744310MetaCfg,
    ArnoldPickupObject35744320MetaCfg,
    ArnoldPickupObject35744330MetaCfg,
    ArnoldPickupObject35744410MetaCfg,
    ArnoldPickupObject35744420MetaCfg,
    ArnoldPickupObject35744430MetaCfg,
    ArnoldPickupObject35744440MetaCfg,
    ArnoldPickupObject35744510MetaCfg,
    ArnoldPickupObject35744520MetaCfg,
    ArnoldPickupObject35744530MetaCfg,
    ArnoldPickupObject35744610MetaCfg,
    ArnoldPickupObject35744620MetaCfg,
    ArnoldPickupObject35744630MetaCfg,
    ArnoldPickupObject35744640MetaCfg,
    ArnoldPickupObject35744710MetaCfg,
    ArnoldPickupObject35744720MetaCfg,
    ArnoldPickupObject35744730MetaCfg,
    ArnoldPickupObject35744740MetaCfg,
    ArnoldPickupObject35744810MetaCfg,
    ArnoldPickupObject35744820MetaCfg,
    ArnoldPickupObject35744830MetaCfg,
    ArnoldPickupObject35744840MetaCfg,
    ArnoldPickupObject35744910MetaCfg,
    ArnoldPickupObject35744920MetaCfg,
    ArnoldPickupObject35744930MetaCfg,
    ArnoldPickupObject35744940MetaCfg,
    ArnoldPickupObject35965020MetaCfg,
    ArnoldPickupObject35965110MetaCfg,
    ArnoldPickupObject35965120MetaCfg,
    ArnoldPickupObject35965130MetaCfg,
    ArnoldPickupObject35965210MetaCfg,
    ArnoldPickupObject35965220MetaCfg,
    ArnoldPickupObject35965230MetaCfg,
    ArnoldPickupObject35965310MetaCfg,
    ArnoldPickupObject35965320MetaCfg,
    ArnoldPickupObject35965330MetaCfg,
    ArnoldPickupObject35965410MetaCfg,
    ArnoldPickupObject35965420MetaCfg,
    ArnoldPickupObject35965430MetaCfg,
    ArnoldPickupObject35965510MetaCfg,
    ArnoldPickupObject35965520MetaCfg,
    ArnoldPickupObject35965530MetaCfg,
    ArnoldPickupObject35965610MetaCfg,
    ArnoldPickupObject35965620MetaCfg,
    ArnoldPickupObject35965630MetaCfg,
    ArnoldPickupObject35965710MetaCfg,
    ArnoldPickupObject35965720MetaCfg,
    ArnoldPickupObject35965730MetaCfg,
    ArnoldPickupObject35965810MetaCfg,
    ArnoldPickupObject35965820MetaCfg,
    ArnoldPickupObject35965830MetaCfg,
)
from .arnold_reorient_object import (
    ArnoldReorientObject385400MetaCfg,
    ArnoldReorientObject385410MetaCfg,
    ArnoldReorientObject385420MetaCfg,
    ArnoldReorientObject385430MetaCfg,
    ArnoldReorientObject385440MetaCfg,
    ArnoldReorientObject385450MetaCfg,
    ArnoldReorientObject385460MetaCfg,
    ArnoldReorientObject385470MetaCfg,
    ArnoldReorientObject385480MetaCfg,
    ArnoldReorientObject393390MetaCfg,
    ArnoldReorientObject3854045MetaCfg,
    ArnoldReorientObject3854145MetaCfg,
    ArnoldReorientObject3854245MetaCfg,
    ArnoldReorientObject3854345MetaCfg,
    ArnoldReorientObject3854445MetaCfg,
    ArnoldReorientObject3854545MetaCfg,
    ArnoldReorientObject3854645MetaCfg,
    ArnoldReorientObject3854745MetaCfg,
    ArnoldReorientObject3854845MetaCfg,
    ArnoldReorientObject3933100MetaCfg,
    ArnoldReorientObject3933110MetaCfg,
    ArnoldReorientObject3933120MetaCfg,
    ArnoldReorientObject3933945MetaCfg,
    ArnoldReorientObject3934130MetaCfg,
    ArnoldReorientObject3934140MetaCfg,
    ArnoldReorientObject3934150MetaCfg,
    ArnoldReorientObject3934160MetaCfg,
    ArnoldReorientObject3934170MetaCfg,
    ArnoldReorientObject3934180MetaCfg,
    ArnoldReorientObject3934190MetaCfg,
    ArnoldReorientObject3990200MetaCfg,
    ArnoldReorientObject3990210MetaCfg,
    ArnoldReorientObject3990220MetaCfg,
    ArnoldReorientObject3990230MetaCfg,
    ArnoldReorientObject3990240MetaCfg,
    ArnoldReorientObject3990250MetaCfg,
    ArnoldReorientObject4084260MetaCfg,
    ArnoldReorientObject4084270MetaCfg,
    ArnoldReorientObject4084280MetaCfg,
    ArnoldReorientObject4084290MetaCfg,
    ArnoldReorientObject4084300MetaCfg,
    ArnoldReorientObject4118310MetaCfg,
    ArnoldReorientObject4118320MetaCfg,
    ArnoldReorientObject4118330MetaCfg,
    ArnoldReorientObject4118340MetaCfg,
    ArnoldReorientObject4118350MetaCfg,
    ArnoldReorientObject4118360MetaCfg,
    ArnoldReorientObject4118370MetaCfg,
    ArnoldReorientObject4118380MetaCfg,
    ArnoldReorientObject4200390MetaCfg,
    ArnoldReorientObject4200400MetaCfg,
    ArnoldReorientObject4200410MetaCfg,
    ArnoldReorientObject4200420MetaCfg,
    ArnoldReorientObject4200430MetaCfg,
    ArnoldReorientObject4233440MetaCfg,
    ArnoldReorientObject4233450MetaCfg,
    ArnoldReorientObject4233460MetaCfg,
    ArnoldReorientObject4233470MetaCfg,
    ArnoldReorientObject4233480MetaCfg,
    ArnoldReorientObject4233490MetaCfg,
    ArnoldReorientObject4233500MetaCfg,
    ArnoldReorientObject4314510MetaCfg,
    ArnoldReorientObject4314520MetaCfg,
    ArnoldReorientObject4314530MetaCfg,
    ArnoldReorientObject4314540MetaCfg,
    ArnoldReorientObject4314550MetaCfg,
    ArnoldReorientObject4393560MetaCfg,
    ArnoldReorientObject4393570MetaCfg,
    ArnoldReorientObject4393580MetaCfg,
    ArnoldReorientObject4393590MetaCfg,
    ArnoldReorientObject38540135MetaCfg,
    ArnoldReorientObject38540180MetaCfg,
    ArnoldReorientObject38541135MetaCfg,
    ArnoldReorientObject38541180MetaCfg,
    ArnoldReorientObject38542135MetaCfg,
    ArnoldReorientObject38542180MetaCfg,
    ArnoldReorientObject38543135MetaCfg,
    ArnoldReorientObject38543180MetaCfg,
    ArnoldReorientObject38544135MetaCfg,
    ArnoldReorientObject38544180MetaCfg,
    ArnoldReorientObject38545135MetaCfg,
    ArnoldReorientObject38545180MetaCfg,
    ArnoldReorientObject38546135MetaCfg,
    ArnoldReorientObject38546180MetaCfg,
    ArnoldReorientObject38547135MetaCfg,
    ArnoldReorientObject38547180MetaCfg,
    ArnoldReorientObject38548135MetaCfg,
    ArnoldReorientObject38548180MetaCfg,
    ArnoldReorientObject39331045MetaCfg,
    ArnoldReorientObject39331145MetaCfg,
    ArnoldReorientObject39331245MetaCfg,
    ArnoldReorientObject39339135MetaCfg,
    ArnoldReorientObject39339180MetaCfg,
    ArnoldReorientObject39341345MetaCfg,
    ArnoldReorientObject39341445MetaCfg,
    ArnoldReorientObject39341545MetaCfg,
    ArnoldReorientObject39341645MetaCfg,
    ArnoldReorientObject39341745MetaCfg,
    ArnoldReorientObject39341845MetaCfg,
    ArnoldReorientObject39341945MetaCfg,
    ArnoldReorientObject39902045MetaCfg,
    ArnoldReorientObject39902145MetaCfg,
    ArnoldReorientObject39902245MetaCfg,
    ArnoldReorientObject39902345MetaCfg,
    ArnoldReorientObject39902445MetaCfg,
    ArnoldReorientObject39902545MetaCfg,
    ArnoldReorientObject40842645MetaCfg,
    ArnoldReorientObject40842745MetaCfg,
    ArnoldReorientObject40842845MetaCfg,
    ArnoldReorientObject40842945MetaCfg,
    ArnoldReorientObject40843045MetaCfg,
    ArnoldReorientObject41183145MetaCfg,
    ArnoldReorientObject41183245MetaCfg,
    ArnoldReorientObject41183345MetaCfg,
    ArnoldReorientObject41183445MetaCfg,
    ArnoldReorientObject41183545MetaCfg,
    ArnoldReorientObject41183645MetaCfg,
    ArnoldReorientObject41183745MetaCfg,
    ArnoldReorientObject41183845MetaCfg,
    ArnoldReorientObject42003945MetaCfg,
    ArnoldReorientObject42004045MetaCfg,
    ArnoldReorientObject42004145MetaCfg,
    ArnoldReorientObject42004245MetaCfg,
    ArnoldReorientObject42004345MetaCfg,
    ArnoldReorientObject42334445MetaCfg,
    ArnoldReorientObject42334545MetaCfg,
    ArnoldReorientObject42334645MetaCfg,
    ArnoldReorientObject42334745MetaCfg,
    ArnoldReorientObject42334845MetaCfg,
    ArnoldReorientObject42334945MetaCfg,
    ArnoldReorientObject42335045MetaCfg,
    ArnoldReorientObject43145145MetaCfg,
    ArnoldReorientObject43145245MetaCfg,
    ArnoldReorientObject43145345MetaCfg,
    ArnoldReorientObject43145445MetaCfg,
    ArnoldReorientObject43145545MetaCfg,
    ArnoldReorientObject43935645MetaCfg,
    ArnoldReorientObject43935745MetaCfg,
    ArnoldReorientObject43935845MetaCfg,
    ArnoldReorientObject43935945MetaCfg,
    ArnoldReorientObject393310135MetaCfg,
    ArnoldReorientObject393310180MetaCfg,
    ArnoldReorientObject393311135MetaCfg,
    ArnoldReorientObject393311180MetaCfg,
    ArnoldReorientObject393312135MetaCfg,
    ArnoldReorientObject393312180MetaCfg,
    ArnoldReorientObject393413180MetaCfg,
    ArnoldReorientObject393414135MetaCfg,
    ArnoldReorientObject393414180MetaCfg,
    ArnoldReorientObject393415135MetaCfg,
    ArnoldReorientObject393415180MetaCfg,
    ArnoldReorientObject393416180MetaCfg,
    ArnoldReorientObject393417135MetaCfg,
    ArnoldReorientObject393417180MetaCfg,
    ArnoldReorientObject393418135MetaCfg,
    ArnoldReorientObject393418180MetaCfg,
    ArnoldReorientObject393419135MetaCfg,
    ArnoldReorientObject393419180MetaCfg,
    ArnoldReorientObject399020135MetaCfg,
    ArnoldReorientObject399020180MetaCfg,
    ArnoldReorientObject399021135MetaCfg,
    ArnoldReorientObject399021180MetaCfg,
    ArnoldReorientObject399022135MetaCfg,
    ArnoldReorientObject399022180MetaCfg,
    ArnoldReorientObject399023135MetaCfg,
    ArnoldReorientObject399023180MetaCfg,
    ArnoldReorientObject399024135MetaCfg,
    ArnoldReorientObject399024180MetaCfg,
    ArnoldReorientObject399025135MetaCfg,
    ArnoldReorientObject399025180MetaCfg,
    ArnoldReorientObject408426180MetaCfg,
    ArnoldReorientObject408427135MetaCfg,
    ArnoldReorientObject408427180MetaCfg,
    ArnoldReorientObject408428135MetaCfg,
    ArnoldReorientObject408428180MetaCfg,
    ArnoldReorientObject408429135MetaCfg,
    ArnoldReorientObject408429180MetaCfg,
    ArnoldReorientObject408430135MetaCfg,
    ArnoldReorientObject408430180MetaCfg,
    ArnoldReorientObject411831180MetaCfg,
    ArnoldReorientObject411832180MetaCfg,
    ArnoldReorientObject411833135MetaCfg,
    ArnoldReorientObject411833180MetaCfg,
    ArnoldReorientObject411834135MetaCfg,
    ArnoldReorientObject411834180MetaCfg,
    ArnoldReorientObject411835135MetaCfg,
    ArnoldReorientObject411835180MetaCfg,
    ArnoldReorientObject411836135MetaCfg,
    ArnoldReorientObject411836180MetaCfg,
    ArnoldReorientObject411837135MetaCfg,
    ArnoldReorientObject411837180MetaCfg,
    ArnoldReorientObject411838135MetaCfg,
    ArnoldReorientObject411838180MetaCfg,
    ArnoldReorientObject420039135MetaCfg,
    ArnoldReorientObject420039180MetaCfg,
    ArnoldReorientObject420040135MetaCfg,
    ArnoldReorientObject420040180MetaCfg,
    ArnoldReorientObject420041135MetaCfg,
    ArnoldReorientObject420041180MetaCfg,
    ArnoldReorientObject420042180MetaCfg,
    ArnoldReorientObject420043180MetaCfg,
    ArnoldReorientObject423344135MetaCfg,
    ArnoldReorientObject423344180MetaCfg,
    ArnoldReorientObject423345180MetaCfg,
    ArnoldReorientObject423346135MetaCfg,
    ArnoldReorientObject423346180MetaCfg,
    ArnoldReorientObject423347135MetaCfg,
    ArnoldReorientObject423347180MetaCfg,
    ArnoldReorientObject423348135MetaCfg,
    ArnoldReorientObject423348180MetaCfg,
    ArnoldReorientObject423349135MetaCfg,
    ArnoldReorientObject423349180MetaCfg,
    ArnoldReorientObject423350135MetaCfg,
    ArnoldReorientObject423350180MetaCfg,
    ArnoldReorientObject431451180MetaCfg,
    ArnoldReorientObject431452180MetaCfg,
    ArnoldReorientObject431453180MetaCfg,
    ArnoldReorientObject431454180MetaCfg,
    ArnoldReorientObject431455180MetaCfg,
    ArnoldReorientObject439356180MetaCfg,
    ArnoldReorientObject439357180MetaCfg,
    ArnoldReorientObject439358180MetaCfg,
    ArnoldReorientObject439359180MetaCfg,
)



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_OBJECTS = [
    RigidObjMetaCfg(
        name="grill_visual",
        usd_path="roboverse_data/assets/rlbench/meat_off_grill/grill_visual/usd/grill_visual.usd",
        physics=PhysicStateType.GEOM,
    ),
    RigidObjMetaCfg(
        name="chicken_visual",
        usd_path="roboverse_data/assets/rlbench/meat_off_grill/chicken_visual/usd/chicken_visual.usd",
        physics=PhysicStateType.RIGIDBODY,
    ),
    RigidObjMetaCfg(
        name="steak_visual",
        usd_path="roboverse_data/assets/rlbench/meat_off_grill/steak_visual/usd/steak_visual.usd",
        physics=PhysicStateType.RIGIDBODY,
    ),
]


@configclass
class MeatOffGrillMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/meat_off_grill/v2"
    objects = _OBJECTS
    # TODO: add checker


@configclass
class MeatOnGrillMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/meat_on_grill/v2"
    objects = _OBJECTS
    # TODO: add checker



"""Configuration classes for lights used in the simulation."""

from __future__ import annotations

import math

import torch

from metasim.utils import configclass
from metasim.utils.math import quat_from_euler_xyz


@configclass
class BaseLightMetaCfg:
    """Base configuration for a light."""

    intensity: float = 500.0
    """Intensity of the light"""
    color: tuple[float, float, float] = (1.0, 1.0, 1.0)
    """Color of the light"""


@configclass
class DistantLightMetaCfg(BaseLightMetaCfg):
    """Configuration for a distant light. The default direction is (0, 0, -1), pointing towards Z- direction."""

    polar: float = 0.0
    """Polar angle of the light (in degrees). Default is 0, which means the light is pointing towards Z- direction."""
    azimuth: float = 0.0
    """Azimuth angle of the light (in degrees). Default is 0."""

    @property
    def quat(self) -> tuple[float, float, float, float]:
        """Quaternion of the light direction. (1, 0, 0, 0) means the light is pointing towards Z- direction."""
        roll = torch.tensor(self.polar / 180.0 * math.pi)
        pitch = torch.tensor(0.0)
        yaw = torch.tensor(self.azimuth / 180.0 * math.pi)
        return tuple(quat_from_euler_xyz(roll, pitch, yaw).squeeze(0).tolist())


@configclass
class CylinderLightMetaCfg(BaseLightMetaCfg):
    """Configuration for a cylinder light."""

    length: float = 1.0
    """Length of the cylinder (in m). Default is 1.0m."""
    radius: float = 0.5
    """Radius of the cylinder (in m). Default is 0.5m."""
    pos: tuple[float, float, float] = (0.0, 0.0, 0.0)
    """Position of the cylinder (in m). Default is (0.0, 0.0, 0.0)."""
    rot: tuple[float, float, float, float] = (1.0, 0.0, 0.0, 0.0)
    """Orientation of the cylinder. Default is (1.0, 0.0, 0.0, 0.0)."""



"""Crawl task for humanoid robots."""

from __future__ import annotations

import torch

from metasim.cfg.checkers import CrawlChecker
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.types import EnvState
from metasim.utils import configclass, humanoid_reward_util
from metasim.utils.humanoid_robot_util import (
    actuator_forces,
    center_of_mass_position,
    center_of_mass_rotation,
    center_of_mass_velocity,
    imu_position,
)

from .base_metacfg import HumanoidTaskMetaCfg


class CrawlReward:
    """Reward function for the crawl task."""

    def __init__(self, robot_name="h1"):
        """Initialize the crawl reward."""
        self.robot_name = robot_name

        # Define the expected quaternion for crawling, normalized
        quat_raw = torch.tensor([0.75, 0, 0.65, 0])
        self.quat_crawl = quat_raw / torch.norm(quat_raw)

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the crawl reward."""
        results = []
        for state in states:
            # Get robot position and IMU data
            robot_pos = center_of_mass_position(state, self.robot_name)
            imu_pos = imu_position(state, self.robot_name)

            # Get pelvis quaternion for orientation calculation
            pelvis_quat = center_of_mass_rotation(state, self.robot_name)

            # Get velocity for speed calculation
            velocity = center_of_mass_velocity(state, self.robot_name)
            vx = velocity[0]

            # heightcrawl = height((0.6,1),1)
            # Use robot height (z position) with tolerance between 0.6 and 1.0
            height_crawl = humanoid_reward_util.tolerance(robot_pos[2], bounds=(0.6, 1.0), margin=1.0)

            # heightIMU = tol(zIMU,(0.6,1),1)
            height_imu = humanoid_reward_util.tolerance(imu_pos[2], bounds=(0.6, 1.0), margin=1.0)

            # orientation = tol(‖quatpelvis−quatcrawl‖,(0,0),1)
            # Calculate quaternion distance and use tolerance function
            quat_diff = torch.norm(pelvis_quat.clone().detach() - self.quat_crawl)
            orientation = humanoid_reward_util.tolerance(quat_diff, bounds=(0, 0), margin=1.0)

            # tunnel = tol(yIMU,(−1,1),0)
            # Reward when IMU's y-coordinate is within the tunnel
            tunnel = humanoid_reward_util.tolerance(imu_pos[1], bounds=(-1, 1), margin=0.0)

            # speed = tol(vx,(1,+∞),1)
            # Reward for forward velocity greater than 1
            speed = humanoid_reward_util.tolerance(vx, bounds=(1, float("inf")), margin=1.0)

            # Calculate the small control factor (e) similar to sit task
            small_control = humanoid_reward_util.tolerance(
                actuator_forces(state, self.robot_name),
                margin=10,
                value_at_margin=0,
                sigmoid="quadratic",
            ).mean()
            small_control = (4 + small_control) / 5

            # R = tunnel × (0.1·e + 0.25·min(heightcrawl,heightIMU) + 0.25·orientation + 0.4·speed)
            min_height = min(height_crawl, height_imu)
            reward = tunnel * (0.1 * small_control + 0.25 * min_height + 0.25 * orientation + 0.4 * speed)

            results.append(reward)

        return torch.tensor(results)


@configclass
class CrawlMetaCfg(HumanoidTaskMetaCfg):
    """Crawl task for humanoid robots."""

    episode_length = 1000
    objects = [
        RigidObjMetaCfg(
            name="tunnel",
            mjcf_path="roboverse_data/assets/humanoidbench/crawl/tunnel/mjcf/tunnel.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/humanoidbench/crawl/v2/initial_state_v2.json"
    checker = CrawlChecker()
    reward_weights = [1.0]
    reward_functions = [CrawlReward]



"""Sub-module containing the base task configuration."""

from __future__ import annotations

from dataclasses import MISSING

import gymnasium as gym
import torch

from metasim.cfg.checkers import BaseChecker
from metasim.cfg.objects import BaseObjMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.types import EnvState
from metasim.utils import configclass


@configclass
class BaseTaskMetaCfg:
    """Base task configuration."""

    decimation: int = 3
    episode_length: int = MISSING
    objects: list[BaseObjMetaCfg] = MISSING
    traj_filepath: str = MISSING
    source_benchmark: BenchmarkType = MISSING
    task_type: TaskType = MISSING
    step_repeat: int = 1  # how many simulation steps to run for each step() call
    checker: BaseChecker = BaseChecker()
    can_tabletop: bool = False
    reward_functions: list[callable[[list[EnvState], str | None], torch.FloatTensor]] = MISSING
    reward_weights: list[float] = MISSING


@configclass
class BaseRLTaskMetaCfg(BaseTaskMetaCfg):
    """Base RL task configuration."""

    # action_space: gym.spaces.Space, can be inferenced from robot (joint_limits)
    observation_space: gym.spaces.Space = MISSING
    observation_function: callable[[list[EnvState]], torch.FloatTensor] = MISSING  # [dummy_obs]

    reward_range: tuple[float, float] = (-float("inf"), float("inf"))



# ruff: noqa: F401

from .assembly_metacfg import AssemblyMetaCfg
from .bin_pick_metacfg import BinPickMetaCfg
from .drawer_open_metacfg import DrawerOpenMetaCfg
from .hammer_metacfg import HammerMetaCfg



from metasim.cfg.objects import ArticulationObjMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_CABINET = ArticulationObjMetaCfg(
    name="cabinet_base",
    usd_path="roboverse_data/assets/rlbench/slide_cabinet_open_and_place_cups/cabinet_base/usd/cabinet_base.usd",
)

_CUP = RigidObjMetaCfg(
    name="cup_visual",
    usd_path="roboverse_data/assets/rlbench/slide_cabinet_open_and_place_cups/cup_visual/usd/cup_visual.usd",
    physics=PhysicStateType.RIGIDBODY,
)


@configclass
class SlideCabinetOpenAndPlaceCupsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/slide_cabinet_open_and_place_cups/v2"
    objects = [_CABINET, _CUP]
    # TODO: add checker


@configclass
class TakeCupOutFromCabinetMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/take_cup_out_from_cabinet/v2"
    objects = [_CABINET, _CUP]
    # TODO: add checker



from __future__ import annotations

#########################################
### Add command line arguments
#########################################
from dataclasses import dataclass

import tyro


@dataclass
class Args:
    src_robot: str
    tgt_robots: list[str]
    tasks: list[str] | None = None
    src_path: str | None = None
    tgt_path: str | None = None
    device: str = "cuda:0"
    ignore_ground_collision: bool = False
    binary_action: bool = False
    """Use binary action for the target robot. If `True`, the resulting action for each step will be either `gripper_release_q` or `gripper_actuate_q`, otherwise the resulting action will be scaled to the new gripper joint limits"""
    viz: bool = False

    def __post_init__(self):
        assert self.tasks is not None or (self.src_path is not None and self.tgt_path is not None), (
            "Either tasks, or both source_path and target_path must be provided"
        )


args = tyro.cli(Args)


#########################################
### Normal code
#########################################
import os

import numpy as np
import rootutils
import torch
from curobo.types.math import Pose
from loguru import logger as log
from rich.logging import RichHandler
from tqdm.rich import tqdm, trange

rootutils.setup_root(__file__, pythonpath=True)
log.configure(handlers=[{"sink": RichHandler(), "format": "{message}"}])

from glob import glob

from metasim.utils.demo_util.loader import load_traj_file, save_traj_file
from metasim.utils.kinematics_utils import ee_pose_from_tcp_pose, get_curobo_models, tcp_pose_from_ee_pose
from metasim.utils.setup_util import get_robot, get_task

NUM_SEED = 20


def single(src_path, tgt_path, src_robot, tgt_robot):
    if args.viz:
        from plotly import graph_objects as go

        from metasim.utils.viz_utils import plot_point_cloud

    # Robot configurations
    src_robot_cfg = get_robot(src_robot)
    tgt_robot_cfg = get_robot(tgt_robot)

    _, src_fk, src_ik = get_curobo_models(src_robot_cfg, args.ignore_ground_collision)
    _, _, tgt_ik = get_curobo_models(tgt_robot_cfg, args.ignore_ground_collision)

    # Prepare file structure
    demo_data = load_traj_file(src_path)

    if tgt_robot == "ur5e_2f85":
        log.warning("Retargetting to robot ur5e_2f85, removing first 30 actions")
        for k, v in demo_data.items():
            for i in range(len(v)):
                demo_data[k][i]["actions"] = demo_data[k][i]["actions"][30:]
                if demo_data[k][i].get("states", None) is not None:
                    demo_data[k][i]["states"] = demo_data[k][i]["states"][30:]

    src_demos = demo_data[src_robot_cfg.name]
    n_demo = len(src_demos)
    episode_lengths = torch.tensor([len(demo["actions"]) for demo in src_demos], dtype=torch.int, device=args.device)
    max_episode_len = episode_lengths.max().item()

    # Compute source trajectory: (JS pose ->) EE pose -> TCP pose
    arm_joint_names = [jn for jn, jn_cfg in src_robot_cfg.actuators.items() if not jn_cfg.is_ee]
    gripper_joint_names = [jn for jn, jn_cfg in src_robot_cfg.actuators.items() if jn_cfg.is_ee]
    curobo_joints = src_ik.robot_config.cspace.joint_names

    robot_pos = np.stack([np.array(demo["init_state"][src_robot_cfg.name]["pos"]) for demo in src_demos], axis=0)
    robot_quat = np.stack([np.array(demo["init_state"][src_robot_cfg.name]["rot"]) for demo in src_demos], axis=0)
    robot_pos = torch.from_numpy(robot_pos).to(args.device).float()
    robot_quat = torch.from_numpy(robot_quat).to(args.device).float()

    robot_q_traj = np.stack(
        [
            np.pad(
                np.asarray([
                    [a["dof_pos_target"][j] for j in (arm_joint_names + gripper_joint_names)] for a in demo["actions"]
                ]),
                [(0, max_episode_len - len(demo["actions"])), (0, 0)],
                mode="edge",
            )
            for demo in src_demos
        ],
        axis=0,
    )

    # Align with cuRobo: Some robot does not have gripper joints
    robot_q_traj = torch.from_numpy(robot_q_traj).to(args.device).float()[..., : len(curobo_joints)]

    tgt_ee_act_rel = np.stack(
        [
            np.pad(
                np.asarray([[a["dof_pos_target"][j] for j in gripper_joint_names] for a in demo["actions"]]),
                [(0, max_episode_len - len(demo["actions"])), (0, 0)],
                mode="edge",
            )
            for demo in src_demos
        ],
        axis=0,
    )
    tgt_ee_act_rel = torch.from_numpy(tgt_ee_act_rel).to(args.device).float()

    if "ee_pose_target" in src_demos[0]["actions"][0]:
        src_ee_pos_traj = np.stack(
            [
                np.pad(
                    np.asarray([a["ee_pose_target"]["pos"] for a in demo["actions"]]),
                    [(0, max_episode_len - len(demo["actions"])), (0, 0)],
                    mode="edge",
                )
                for demo in src_demos
            ],
            axis=0,
        )
        src_ee_quat_traj = np.stack(
            [
                np.pad(
                    np.asarray([a["ee_pose_target"]["rot"] for a in demo["actions"]]),
                    [(0, max_episode_len - len(demo["actions"])), (0, 0)],
                    mode="edge",
                )
                for demo in src_demos
            ],
            axis=0,
        )
        tgt_ee_act_rel = np.stack(
            [
                np.pad(
                    np.array([a["ee_pose_target"]["gripper_joint_pos"] for a in demo["actions"]]),
                    [(0, max_episode_len - len(demo["actions"])), (0, 0)],
                    mode="edge",
                )
                for demo in src_demos
            ],
            axis=0,
        )
        src_ee_pos_traj = torch.from_numpy(src_ee_pos_traj).to(args.device).float()
        src_ee_quat_traj = torch.from_numpy(src_ee_quat_traj).to(args.device).float()
    else:
        # The EE information are missing, compute from JS pose
        log.warning("EE pose information is missing, computing from joint space pose")
        src_ee_pos_traj, src_ee_quat_traj = [], []
        for i_traj in trange(len(src_demos), desc=f"Computing EE pose for source robot: {src_robot_cfg.name}"):
            # JS pose -> EE pose (local) -> EE pose (world)
            ee_traj, quat_traj = src_fk(robot_q_traj[i_traj].contiguous())
            src_ee_pos_traj.append(ee_traj.clone())
            src_ee_quat_traj.append(quat_traj.clone())

        # These are still in the robot's local frame (for solving IK)
        src_ee_pos_traj = torch.stack(src_ee_pos_traj, dim=0)
        src_ee_quat_traj = torch.stack(src_ee_quat_traj, dim=0)

    tcp_pos_traj, tcp_quat_traj = tcp_pose_from_ee_pose(src_robot_cfg, src_ee_pos_traj, src_ee_quat_traj)

    tcp_pos_traj = tcp_pos_traj.reshape(n_demo, max_episode_len, 3)
    tcp_quat_traj = tcp_quat_traj.reshape(n_demo, max_episode_len, 4)

    # Visualziation
    to_plot = []
    if args.viz:
        to_plot.append(
            plot_point_cloud(
                src_ee_pos_traj[0].cpu().numpy(),
                name=f"Source: {src_robot_cfg.name}",
                marker=dict(color=tgt_ee_act_rel[0].cpu().numpy()),
            )
        )

    # For target embodiment, compute target trajecotry via IK: TCP pose -> EE pose -> JS pose
    succ = torch.ones([n_demo], dtype=torch.bool, device=args.device)

    target_dof = len(tgt_ik.robot_config.cspace.joint_names)
    tgt_arm_qpos = torch.zeros([n_demo, max_episode_len, target_dof], device=args.device)

    tgt_ee_pos_traj = torch.zeros([n_demo, max_episode_len, 3], device=args.device)
    tgt_ee_quat_traj = torch.zeros([n_demo, max_episode_len, 4], device=args.device)

    log.info(f"Retargetting: {src_robot_cfg.name} -> {tgt_robot_cfg.name}")
    tt = trange(max_episode_len, desc=f"{src_robot_cfg.name} -> {tgt_robot_cfg.name}")
    for t in tt:
        # Only solve IK for unfinished and unfailed trajectories
        valid_idx = torch.where((t <= episode_lengths) | succ)[0]
        if len(valid_idx) == 0:
            log.error("gg")
            break

        n_succ, n_fail = ((t >= episode_lengths - 1) | succ).sum(), (~succ).sum()

        tgt_ee_pos_traj[valid_idx, t], tgt_ee_quat_traj[valid_idx, t] = ee_pose_from_tcp_pose(
            tgt_robot_cfg, tcp_pos_traj[valid_idx, t], tcp_quat_traj[valid_idx, t]
        )

        if t > 0:
            seed_config = tgt_arm_qpos[:, t - 1 : t].tile([1, NUM_SEED, 1])
        else:
            seed_config = torch.zeros_like(tgt_arm_qpos[:, t - 1 : t]).tile([1, NUM_SEED, 1])
            seed_config[:, :, 1] = -20
        result = tgt_ik.solve_batch(Pose(tgt_ee_pos_traj[:, t], tgt_ee_quat_traj[:, t]), seed_config=seed_config)

        tgt_arm_qpos[valid_idx, t] = result.solution[valid_idx, 0]
        succ[valid_idx] = succ[valid_idx] * result.success.squeeze()[valid_idx]

        tt.set_description(
            f"{src_robot_cfg.name} -> {tgt_robot_cfg.name} | {n_succ.item()!s} succ | {n_fail.item()!s} fail | {n_demo!s} total"
        )
        tt.refresh()

    if args.viz:
        to_plot.append(plot_point_cloud(tgt_ee_pos_traj[0].cpu().numpy(), name=tgt_robot_cfg.name))

    src_release_q = torch.tensor(src_robot_cfg.gripper_release_q, device=args.device)[None, None, :]
    src_actuate_q = torch.tensor(src_robot_cfg.gripper_actuate_q, device=args.device)[None, None, :]
    tgt_release_q = torch.tensor(tgt_robot_cfg.gripper_release_q, device=args.device)[None, None, :]
    tgt_actuate_q = torch.tensor(tgt_robot_cfg.gripper_actuate_q, device=args.device)[None, None, :]
    src_ee_act_rel = torch.abs(tgt_ee_act_rel - src_release_q) / torch.abs(src_actuate_q - src_release_q)
    src_ee_act_rel = src_ee_act_rel.mean(axis=-1)  # Avg over two fingers

    if args.binary_action:
        src_ee_act_rel = (src_ee_act_rel > 0.5).float()
    tgt_ee_act_rel = src_ee_act_rel[:, :, None] * tgt_actuate_q + (1 - src_ee_act_rel[:, :, None]) * tgt_release_q

    tgt_robot_qpos = torch.cat([tgt_arm_qpos, tgt_ee_act_rel], dim=-1)
    tgt_robot_joint_names = list(tgt_robot_cfg.actuators.keys())

    log.info(f"Success rate: {(succ.sum().item() / n_demo * 100):.2f}%, saving trajectories...")
    tgt_demos = []
    for i_succ in tqdm(torch.where(succ)[0], desc=f"Saving trajectories for {tgt_robot_cfg.name}"):
        demo = {"actions": [], "states": [], "init_state": {}}
        for i_step in range(episode_lengths[i_succ]):
            demo["actions"].append({
                "dof_pos_target": dict(zip(tgt_robot_joint_names, tgt_robot_qpos[i_succ, i_step].tolist()))
            })
        demo["init_state"] = src_demos[i_succ]["init_state"]
        del demo["init_state"][src_robot_cfg.name]
        demo["init_state"][tgt_robot_cfg.name] = {
            "dof_pos": dict(zip(tgt_robot_joint_names, tgt_robot_qpos[i_succ, 0].tolist())),
            "pos": robot_pos[i_succ].tolist(),
            "rot": robot_quat[i_succ].tolist(),
        }
        tgt_demos.append(demo)
    log.info(f"Saving trajectories for {tgt_robot_cfg.name}")

    if args.viz:
        go.Figure(to_plot).show()

    if not os.path.exists(os.path.dirname(tgt_path)):
        os.makedirs(os.path.dirname(tgt_path), exist_ok=True)
    save_traj_file({tgt_robot_cfg.name: tgt_demos}, tgt_path)
    log.info(f"Retarget finished -> {tgt_path}")


def main():
    if args.tasks is not None:
        for task_name in args.tasks:
            task = get_task(task_name)
            src_path = task.traj_filepath
            if os.path.isdir(src_path):
                paths = glob(os.path.join(src_path, f"{args.src_robot}_v2.*"))
                assert len(paths) >= 1, f"No trajectories found for {args.src_robot}"
                src_path = paths[0]
            for tgt_robot in args.tgt_robots:
                tgt_path = os.path.join(os.path.dirname(src_path), f"{tgt_robot}_v2.pkl.gz")
                single(src_path, tgt_path, args.src_robot, tgt_robot)
    else:
        single(args.src_path, args.tgt_path)


if __name__ == "__main__":
    main()



"""Sub-module containing the scenario configuration."""

from __future__ import annotations

import os
from typing import Literal

from loguru import logger as log
from tyro import MISSING

from metasim.utils.configclass import configclass
from metasim.utils.io_util import check_and_download
from metasim.utils.setup_util import get_robot, get_scene, get_task

from .cameras.camera import PinholeCameraMetaCfg
from .checkers import BaseChecker, EmptyChecker
from .lights import BaseLightMetaCfg, CylinderLightMetaCfg, DistantLightMetaCfg
from .objects import BaseObjMetaCfg, PrimitiveCubeMetaCfg, PrimitiveSphereMetaCfg
from .randomization import RandomizationMetaCfg
from .render import RenderMetaCfg
from .robots.base_robot_metacfg import BaseRobotMetaCfg
from .scenes.base_scene_metacfg import SceneMetaCfg
from .tasks.base_task_metacfg import BaseTaskMetaCfg


def check_asset(obj: BaseObjMetaCfg, sim: Literal["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien", "mujoco"]):
    """Check and download the asset."""
    ## TODO: add a primitive base class?
    if isinstance(obj, PrimitiveCubeMetaCfg) or isinstance(obj, PrimitiveSphereMetaCfg):
        return

    if sim in ["isaaclab"]:
        check_and_download(obj.usd_path)
    elif sim in ["isaacgym", "pybullet", "sapien"]:
        check_and_download(obj.urdf_path)
    elif sim in ["mujoco"]:
        check_and_download(obj.mjcf_path)


@configclass
class ScenarioMetaCfg:
    """Scenario configuration."""

    task: BaseTaskMetaCfg | None = None  # This item should be removed?
    """None means no task specified"""
    robot: BaseRobotMetaCfg = MISSING
    scene: SceneMetaCfg | None = None
    """None means no scene"""
    lights: list[BaseLightMetaCfg] = [DistantLightMetaCfg()]
    objects: list[BaseObjMetaCfg] = []
    cameras: list[PinholeCameraMetaCfg] = []
    checker: BaseChecker = EmptyChecker()
    render: RenderMetaCfg = RenderMetaCfg()
    random: RandomizationMetaCfg = RandomizationMetaCfg()

    ## Handlers
    sim: Literal["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien", "mujoco"] = "isaaclab"
    renderer: Literal["isaaclab", "isaacgym", "pyrep", "pybullet", "sapien", "mujoco"] | None = None

    ## Others
    num_envs: int = 1
    decimation: int = 1
    episode_length: int = 10000000  # never timeout
    try_add_table: bool = True
    object_states: bool = False
    split: Literal["train", "val", "test", "all"] = "all"
    headless: bool = False

    def __post_init__(self):
        """Post-initialization configuration."""
        ### Align configurations
        if (self.random.scene or self.scene is not None) and self.try_add_table:
            log.warning("try_add_table is set to False because scene randomization is enabled or a scene is specified")
            self.try_add_table = False

        ### Randomization level configuration
        if self.random.level >= 1:
            self.lights = [
                CylinderLightMetaCfg(pos=(-2.0, -2.0, 4.0), intensity=5e4),
                CylinderLightMetaCfg(pos=(-2.0, 2.0, 4.0), intensity=5e4),
                CylinderLightMetaCfg(pos=(2.0, -2.0, 4.0), intensity=5e4),
                CylinderLightMetaCfg(pos=(2.0, 2.0, 4.0), intensity=5e4),
            ]

        ### Parse task and robot
        if isinstance(self.task, str):
            self.task = get_task(self.task)
        if isinstance(self.robot, str):
            self.robot = get_robot(self.robot)
        if isinstance(self.scene, str):
            self.scene = get_scene(self.scene)

        ### Check and download all the paths
        ## Object paths
        objects = self.task.objects if self.task is not None else self.objects
        for obj in objects:
            check_asset(obj, self.sim)
        ## Robot paths
        check_asset(self.robot, self.sim)
        ## Scene paths
        if self.scene is not None:
            check_asset(self.scene, self.sim)
        ## Traj paths
        if self.task is not None:
            traj_filepath = self.task.traj_filepath
            if traj_filepath is None:
                return
            if (
                traj_filepath.find(".pkl") == -1
                and traj_filepath.find(".json") == -1
                and traj_filepath.find(".yaml") == -1
                and traj_filepath.find(".yml") == -1
            ):
                if os.path.exists(os.path.join(traj_filepath, f"{self.robot.name}_v2.pkl.gz")):
                    traj_filepath = os.path.join(traj_filepath, f"{self.robot.name}_v2.pkl.gz")
                else:
                    traj_filepath = os.path.join(traj_filepath, f"{self.robot.name}_v2.pkl")
            check_and_download(traj_filepath)



from metasim.cfg.tasks import BaseTaskMetaCfg
from metasim.constants import BenchmarkType, TaskType
from metasim.utils import configclass


@configclass
class DexterousHandsTaskMetaCfg(BaseTaskMetaCfg):
    """Dexterous Hands Task Meta Configuration."""

    source_benchmark = BenchmarkType.DEXTEROUS_HANDS
    task_type = TaskType.DEXTEROUS_HANDS
    can_tabletop = False



# ruff: noqa: F401

from .libero_objects.libero_pick_alphabet_soup import LiberoPickAlphabetSoupMetaCfg
from .libero_objects.libero_pick_bbq_sauce import LiberoPickBbqSauceMetaCfg
from .libero_objects.libero_pick_butter import LiberoPickButterMetaCfg
from .libero_objects.libero_pick_chocolate_pudding import LiberoPickChocolatePuddingMetaCfg
from .libero_objects.libero_pick_cream_cheese import LiberoPickCreamCheeseMetaCfg
from .libero_objects.libero_pick_ketchup import LiberoPickKetchupMetaCfg
from .libero_objects.libero_pick_milk import LiberoPickMilkMetaCfg
from .libero_objects.libero_pick_orange_juice import LiberoPickOrangeJuiceMetaCfg
from .libero_objects.libero_pick_salad_dressing import LiberoPickSaladDressingMetaCfg
from .libero_objects.libero_pick_tomato_sauce import LiberoPickTomatoSauceMetaCfg



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class StackChairsMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/stack_chairs/v2"
    objects = [
        RigidObjMetaCfg(
            name="chair1",
            usd_path="roboverse_data/assets/rlbench/stack_chairs/chair1/usd/chair1.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="chair2",
            usd_path="roboverse_data/assets/rlbench/stack_chairs/chair2/usd/chair2.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="chair3",
            usd_path="roboverse_data/assets/rlbench/stack_chairs/chair3/usd/chair3.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class PutUmbrellaInUmbrellaStandMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/put_umbrella_in_umbrella_stand/v2"
    objects = [
        RigidObjMetaCfg(
            name="umbrella_visual",
            usd_path="roboverse_data/assets/rlbench/put_umbrella_in_umbrella_stand/umbrella_visual/usd/umbrella_visual.usd",
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="stand_visual",
            usd_path="roboverse_data/assets/rlbench/put_umbrella_in_umbrella_stand/stand_visual/usd/stand_visual.usd",
            physics=PhysicStateType.GEOM,
        ),
    ]
    # TODO: add checker



from metasim.cfg.objects import PrimitiveCubeMetaCfg, RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg


@configclass
class ReachAndDragMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/reach_and_drag/v2"
    objects = [
        PrimitiveCubeMetaCfg(
            name="cube",
            size=[0.08, 0.08, 0.08],
            color=[0.85, 0.85, 1.0],
            physics=PhysicStateType.RIGIDBODY,
        ),
        PrimitiveCubeMetaCfg(
            name="stick",
            size=[0.0128, 0.0128, 0.36],
            color=[0.85, 0.85, 1.0],
            physics=PhysicStateType.RIGIDBODY,
        ),
        RigidObjMetaCfg(
            name="target0",
            usd_path="roboverse_data/assets/rlbench/reach_and_drag/target0/usd/target0.usd",
            physics=PhysicStateType.GEOM,
        ),
    ]
    # TODO: add checker



"""Configuration classes for various types of objects used in the simulation.

Configurations include the object name, source file, geometries (scaling, radius, etc.),
and physics (mass, density, etc.).
"""

from __future__ import annotations

import math
from dataclasses import MISSING

from metasim.constants import PhysicStateType
from metasim.utils import configclass


@configclass
class BaseObjMetaCfg:
    """Base class for object metacfg."""

    name: str = MISSING
    """Object name"""
    fix_base_link: bool = False
    """Whether to fix the base link of the object, default is False"""
    scale: float = 1.0
    """Object scaling (in scalar) for the object, default is 1.0"""


# File-based object
@configclass
class RigidObjMetaCfg(BaseObjMetaCfg):
    """Rigid object metacfg.

    The file source should be specified, from a USD, URDF, MJCF, or mesh file,
    with the path specified by the members below.

    Attributes:
        usd_path: Path to the USD file
        urdf_path: Path to the URDF file
        mjcf_path: Path to the MJCF file
        mesh_path: Path to the Mesh file
        physics: Specify the physics APIs applied on the object
    """

    usd_path: str | None = None
    urdf_path: str | None = None
    mjcf_path: str | None = None
    mesh_path: str | None = None
    physics: PhysicStateType = MISSING


@configclass
class NonConvexRigidObjMetaCfg(RigidObjMetaCfg):
    """Non-convex rigid object class."""

    mesh_pose: list[float] = MISSING


@configclass
class ArticulationObjMetaCfg(BaseObjMetaCfg):
    """Articulation object metacfg."""

    usd_path: str | None = None
    urdf_path: str | None = None
    mjcf_path: str | None = None


# Primitive object are all rigid objects
@configclass
class PrimitiveCubeMetaCfg(RigidObjMetaCfg):
    """Primitive cube object metacfg.

    This class specifies configuration parameters of a primitive cube.

    Attributes:
        mass: Mass of the object, in kg, default is 0.1
        color: Color of the object in RGB
        size: Size of the object, extent in m
    """

    mass: float = 0.1
    color: list[float] = MISSING
    size: list[float] = MISSING
    physics: PhysicStateType = MISSING
    mjcf_path: str | None = None  # TODO: remove this field

    @property
    def half_size(self) -> list[float]:
        """Half of the extend, for SAPIEN usage."""
        return [size / 2 for size in self.size]

    @property
    def density(self) -> float:
        """Object density, for SAPIEN usage."""
        return self.mass / (self.size[0] * self.size[1] * self.size[2])


@configclass
class PrimitiveSphereMetaCfg(RigidObjMetaCfg):
    """Primitive sphere object metacfg."""

    mass: float = 0.1
    color: list[float] = MISSING
    radius: float = MISSING
    physics: PhysicStateType = MISSING

    @property
    def density(self) -> float:
        """For SAPIEN usage."""
        return self.mass / (4 / 3 * math.pi * self.radius**3)


@configclass
class PrimitiveCylinderMetaCfg(RigidObjMetaCfg):
    """Primitive cylinder object metacfg."""

    mass: float = 0.1
    color: list[float] = MISSING
    radius: float = MISSING
    height: float = MISSING
    physics: PhysicStateType = MISSING

    @property
    def density(self) -> float:
        """For SAPIEN usage."""
        return self.mass / (math.pi * self.radius**2 * self.height)



from metasim.cfg.objects import PrimitiveCylinderMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg
from .setup_chess_metacfg import _CHESS_BOARD

_WHITE_CHECKERS = [
    PrimitiveCylinderMetaCfg(
        name=f"checker{idx}",
        radius=0.017945,
        height=0.00718,
        color=[1.0, 1.0, 1.0],
        physics=PhysicStateType.RIGIDBODY,
    )
    for idx in range(12)
]

_RED_CHECKERS = [
    PrimitiveCylinderMetaCfg(
        name=f"checker{idx}",
        radius=0.017945,
        height=0.00718,
        color=[1.0, 0.0, 0.0],
        physics=PhysicStateType.RIGIDBODY,
    )
    for idx in range(12, 24)
]


@configclass
class SetupCheckersMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/setup_checkers/v2"
    objects = [_CHESS_BOARD] + _WHITE_CHECKERS + _RED_CHECKERS
    # TODO: add checker



"""Rotate Cube in hand task for humanoid robots."""

from __future__ import annotations

import torch

from metasim.cfg.checkers import CubeChecker
from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.types import EnvState
from metasim.utils import configclass, humanoid_reward_util, humanoid_robot_util

from .base_metacfg import HumanoidTaskMetaCfg, StableReward


class StandingReward:
    """Reward function for maintaining standing posture."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the standing reward."""
        self._robot_name = robot_name
        self._stand_height = 0.6  # 需要根据实际机器人调整

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the standing reward."""
        results_still = []
        for state in states:
            com_vel = humanoid_robot_util.center_of_mass_velocity(state, self._robot_name)
            still_x = humanoid_reward_util.tolerance(com_vel[0], bounds=(0.0, 0.0), margin=2)
            still_y = humanoid_reward_util.tolerance(com_vel[1], bounds=(0.0, 0.0), margin=2)
            still_reward = (still_x + still_y) / 2
            results_still.append(still_reward)

        stable_rewards = StableReward(robot_name=self._robot_name)(states)
        return torch.tensor(results_still) * stable_rewards


class OrientationReward:
    """Reward function for cube orientation alignment."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the orientation reward."""
        self._robot_name = robot_name

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the orientation reward."""
        results = []
        for state in states:
            left_cube_rot = state["metasim_body_cube_1/cube_1"]["rot"]
            right_cube_rot = state["metasim_body_cube_2/cube_2"]["rot"]
            target_cube_rot = state["metasim_body_cube_destination/cube_destination"]["rot"]

            left_alignment = torch.norm(left_cube_rot - target_cube_rot)
            right_alignment = torch.norm(right_cube_rot - target_cube_rot)

            results.append(left_alignment + right_alignment)
        return torch.tensor(results)


class HandProximityReward:
    """Reward function for hand-cube proximity."""

    def __init__(self, robot_name="h1_simple_hand"):
        """Initialize the hand proximity reward."""
        self._robot_name = robot_name

    def __call__(self, states: list[EnvState]) -> torch.FloatTensor:
        """Compute the hand proximity reward."""
        results = []
        for state in states:
            left_hand_pos = humanoid_robot_util.left_hand_position(state, self._robot_name)
            right_hand_pos = humanoid_robot_util.right_hand_position(state, self._robot_name)
            cube1_pos = state["metasim_body_cube_1/cube_1"]["pos"]
            cube2_pos = state["metasim_body_cube_2/cube_2"]["pos"]

            left_dist = torch.norm(left_hand_pos - cube1_pos)
            right_dist = torch.norm(right_hand_pos - cube2_pos)

            left_proximity = humanoid_reward_util.tolerance(left_dist, bounds=(0.0, 0.0), margin=0.5)
            right_proximity = humanoid_reward_util.tolerance(right_dist, bounds=(0.0, 0.0), margin=0.5)

            results.append((left_proximity + right_proximity) / 2)
        return torch.tensor(results)


@configclass
class CubeMetaCfg(HumanoidTaskMetaCfg):
    """Cube task for humanoid robots."""

    episode_length = 1000
    objects = [
        RigidObjMetaCfg(
            name="cube_1",
            mjcf_path="roboverse_data/assets/humanoidbench/cube/cube_1/mjcf/cube_1.xml",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="cube_2",
            mjcf_path="roboverse_data/assets/humanoidbench/cube/cube_2/mjcf/cube_2.xml",
            physics=PhysicStateType.GEOM,
        ),
        RigidObjMetaCfg(
            name="cube_destination",
            mjcf_path="roboverse_data/assets/humanoidbench/cube/cube_destination/mjcf/cube_destination.xml",
            physics=PhysicStateType.GEOM,
            fix_base_link=True,
        ),
    ]
    traj_filepath = "roboverse_data/trajs/humanoidbench/cube/v2/initial_state_v2.json"
    checker = CubeChecker()
    reward_weights = [0.2, 0.5, 0.3]
    reward_functions = [StandingReward, OrientationReward, HandProximityReward]



from metasim.cfg.objects import RigidObjMetaCfg
from metasim.constants import PhysicStateType
from metasim.utils import configclass

from .rlbench_task_metacfg import RLBenchTaskMetaCfg

_PAWNS = [
    RigidObjMetaCfg(
        name=f"{color}_pawn_{idx}",
        usd_path=f"roboverse_data/assets/rlbench/setup_chess/{color}_pawn_a/usd/{color}_pawn_a.usd",  # reuse the same asset for all pawns
        physics=PhysicStateType.RIGIDBODY,
    )
    for color in ["white", "black"]
    for idx in "abcdefgh"
]

_KINGS_AND_QUEENS = [
    RigidObjMetaCfg(
        name=f"{color}_{piece}",
        usd_path=f"roboverse_data/assets/rlbench/setup_chess/{color}_{piece}/usd/{color}_{piece}.usd",
        physics=PhysicStateType.RIGIDBODY,
    )
    for color in ["white", "black"]
    for piece in ["king", "queen"]
]

_OTHER_PIECES = [
    RigidObjMetaCfg(
        name=f"{color}_{side}_{piece}",
        usd_path=f"roboverse_data/assets/rlbench/setup_chess/{color}_kingside_{piece}/usd/{color}_kingside_{piece}.usd",  # reuse the same asset for both sides
        physics=PhysicStateType.RIGIDBODY,
    )
    for color in ["white", "black"]
    for piece in ["rook", "knight", "bishop"]
    for side in ["kingside", "queenside"]
]

_CHESS_BOARD = RigidObjMetaCfg(
    name="chess_board_base_visual",
    usd_path="roboverse_data/assets/rlbench/setup_chess/chess_board_base_visual/usd/chess_board_base_visual.usd",
    physics=PhysicStateType.GEOM,
)


@configclass
class SetupChessMetaCfg(RLBenchTaskMetaCfg):
    episode_length = 200
    traj_filepath = "roboverse_data/trajs/rlbench/setup_chess/v2"
    objects = [_CHESS_BOARD] + _PAWNS + _KINGS_AND_QUEENS + _OTHER_PIECES
    # TODO: add checker
